# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import timedelta


from requests import (
    Request,
    Session,
)


def get_address(charm, short):
    if short:
        return "cs:%s/%s" % (charm["series"], charm["name"])
    else:
        return "cs:~%s/%s/%s" % (
            charm["owner"], charm["series"], charm["name"])


def make_store_url(revision, address=None):
    """Make a store URL given a revision, store address or charm."""
    return address + "-%d" % revision


class CharmStore:
    """Class for accessing the charm store."""

    STAT_URL = "https://store.juju.ubuntu.com/stats/counter"
    STORE_URL = "https://store.juju.ubuntu.com"

    def __init__(self):
        self.session = Session()

    @staticmethod
    def date_str(date):
        """Return a date string as used in charm store query params.

        Unlike date.isoformat, this works consistently for dates and
        datetimes.
        """
        return date.strftime('%Y-%m-%d')

    @classmethod
    def get_stats_range_request(cls, charm, start=None, end=None):
        """Produce a Request for retrieving download counts in a range.

        :param charm: The charm to get information about.
        :param start: The beginning of the date range (as a datetime.date)
        :param end: The end of the date range (as a datetime.date)
        """
        elements = ['charm-bundle', charm['series'], charm['name']]
        if charm['store_url'].startswith('cs:~'):
            elements.append(charm['owner'])
        params = {
            'format': 'json',
        }
        if (start and end):
            params.update({
                'start': cls.date_str(start),
                'end': cls.date_str(end),
                'by': 'day',
            })

        return Request(
            'GET', '%s/%s' % (cls.STAT_URL, ':'.join(elements)),
            params=params
        )

    @classmethod
    def get_charms_info_request(cls, addresses):
        return Request('POST', cls.STORE_URL + '/charm-info',
                       data={'stats': '0', 'charms': addresses})

    def get_charms_info(self, addresses):
        request = self.get_charms_info_request(addresses)
        response = self.session.send(request.prepare())
        return response.json()

    def get_download_counts(self, charm, start=None, end=None):
        """Retrieve download counts for each day in the specified range.

        :param charm: The charm to get information about.
        :param start: The beginning of the date range (as a datetime.date)
        :param end: The end of the date range (as a datetime.date)
        """
        request = self.get_stats_range_request(charm, start, end)
        result = self.session.send(request.prepare())
        json_result = result.json()
        if json_result == [[0]]:
            return []
        return json_result

    def count_downloads_in_days(self, charm, days, end):
        """Return the number of downloads in a given number of days.

        This will return 0 for charms that do not exist.  Presumably the charm
        store does this because charms that do not exist cannot have been
        downloaded.

        :param charm: The charm to get information about.
        :param days: The number of days to look back.
        :param end: The end of the date range (as a datetime.date).
        """
        start = end - timedelta(days=days)
        counts = self.get_download_counts(charm, start, end)
        return sum(downloads for date, downloads in counts)


def get_store_charm_info(store, charms):
    addresses = []
    charm_addresses = []
    for charm in charms:
        possible_addresses = {'long': get_address(charm, False)}
        if charm['promulgated']:
            possible_addresses['short'] = get_address(charm, True)
        charm_addresses.append(possible_addresses)
        addresses.extend(possible_addresses.values())
    info = store.get_charms_info(addresses)
    result = []
    for possible_addresses in charm_addresses:
        if ('short' in possible_addresses and
            'errors' not in info[possible_addresses['short']]):
            address = possible_addresses['short']
        else:
            address = possible_addresses['long']
        result.append((address, info[address]))
    return result
