# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

import argparse
from datetime import (
    datetime,
)
import logging
import pymongo

from charmworld.charmstore import (
    CharmStore,
    get_store_charm_info,
)
from charmworld.jobs.ingest import update_from_store
from charmworld.lp import (
    BASKET_SERIES,
    get_branch_tips,
)
from charmworld.models import (
    getconnection,
    getdb,
)
from charmworld.utils import (
    configure_logging,
    get_ini,
)
from config import (
    BASKET_QUEUE,
    CHARM_IMPORT_FILTER,
    CHARM_QUEUE,
)
from utils import get_queue
from utils import lock
from utils import LockHeld
from utils import parse_branch


default = object()


def all_branch_data(charm_data=None):
    # Return data needed for the ingest job for charm branches on Launchpad.
    log = logging.getLogger("charm.launchpad")
    charms = []
    baskets = []
    if not charm_data:
        charm_data = get_branch_tips()

    for repo, commit, series in charm_data:
        _, branch_name = repo.rsplit("/", 1)

        data = parse_branch(repo, series, commit)

        data['promulgated'] = data['series'] in data['distro_series']

        if data['series'] == BASKET_SERIES:
            if data['bname'] == 'bundle':
                baskets.append(data)
            continue

        if data["bname"] != "trunk":
            log.debug("Skipped branch %s", repo)
            continue

        data['branch_deleted'] = False
        charms.append(data)
    return charms, baskets


def db_charms(db):
    # Return data needed for the ingest job for charms stored in
    # charmworld's database.
    for charm in db.charms.find(
            fields=('owner', 'series', 'name', 'bname', 'branch_spec'),
            sort=[('branch_spec', pymongo.ASCENDING)]):
        # Override the promulgation related settings: Data from this
        # iterator is used by all_charms() only when the charm has no
        # longer a related Launchpad branch. A charm is promulgated if
        # its LP branch is linked to a sourcepackage, hence a charm
        # without a branch cannot be promulgated.
        charm['promulgated'] = False
        charm['distro_series'] = []
        charm['branch_deleted'] = True
        yield charm


def all_charms(db, available, limit=None, import_filter=default):
    # Return a sequence of charms that have a Launchpad branch or that are
    # already known in charmworld's database.
    log = logging.getLogger("charm.launchpad")
    if import_filter is default:
        import_filter = CHARM_IMPORT_FILTER
    all_ = dict((charm['branch_spec'], charm) for charm in db_charms(db))
    all_.update((charm['branch_spec'], charm) for charm in available)
    all_ = sorted(all_.values(), key=lambda charm: charm['branch_spec'])

    count = 0
    for charm in all_:
        if import_filter and not charm['branch_spec'].startswith(
                import_filter):
            continue
        log.info("Queueing charm %s", charm['branch_spec'])
        yield charm
        count += 1
        if limit and count >= limit:
            log.info("Import limit reached (%d)... stopping", limit)
            break


def queue_from_branches(db, store, charm_queue, basket_queue, limit=None,
                        import_filter=default, check_time=None):
    log = logging.getLogger("charm.launchpad")
    charm_data, baskets = all_branch_data()
    charms = list(all_charms(db, charm_data, limit=limit,
                             import_filter=import_filter))
    charms_info = get_store_charm_info(store, charms)
    if check_time is None:
        check_time = datetime.now().ctime()
    for charm, (address, store_data) in zip(charms, charms_info):
        update_from_store(charm, address, store_data, check_time, log)
        added = charm_queue.put(charm)
        if added and 0:
            log.info("Queued %s", charm)
    for basket in baskets:
        log.info("Queueing bundle %s", basket['branch_spec'])
        basket_queue.put(basket)


def dequeue():
    """Dequeue all the charm data in the CHARM_QUEUE.

    This is a script entry point to dequeue all charms. This might be done when
    when a running instance will get a code update that is incompatible with
    charm data in the queue.
    """
    configure_logging()
    log = logging.getLogger("charm.launchpad")
    settings = get_ini()
    connection = getconnection(settings)
    db = getdb(connection, settings.get('mongo.database'))
    try:
        lease_time = int(settings['script_lease_time']) * 60
        with lock('ingest-queue', lease_time, db, log):
            queue = get_queue(db, CHARM_QUEUE)
            queue_size = queue.size()
            queue.clear()
            log.info("Dequeued %s charms", queue_size)
    except LockHeld, error:
        log.warn(str(error))


# XXX bug=1130732: This function does not appear to be tested.
def main():
    configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--limit",
        help="Maximum number of charms to queue.",
        type=int)
    parser.add_argument(
        "--prefix",
        help="Only charms matching this prefix are queued.",
        type=str, default=default)
    args = parser.parse_args()
    log = logging.getLogger("charm.launchpad")
    settings = get_ini()
    if args.limit is None:
        charm_import_limit = settings.get('charm_import_limit')
        charm_import_limit = (
            int(charm_import_limit) if charm_import_limit else None)
    else:
        charm_import_limit = args.limit
    connection = getconnection(settings)
    db = getdb(connection, settings.get('mongo.database'))
    try:
        with lock('ingest-queue', int(settings['script_lease_time']) * 60,
                  db, log):
            charm_queue = get_queue(db, CHARM_QUEUE)
            basket_queue = get_queue(db, BASKET_QUEUE)
            queue_from_branches(db, CharmStore(), charm_queue, basket_queue,
                                charm_import_limit, args.prefix)
    except LockHeld, error:
        log.warn(str(error))


if __name__ == '__main__':
    main()
