# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

"""
Charm review queue

 Contributed by Juan Negronil


 Changes
 - convert to job.
 - use dateutil date parsing instead of string comparsion.
 - merge proposals show target with series/branch.

"""

from launchpadlib.launchpad import Launchpad

import itertools
import pymongo
import logging

from charmworld.utils import (
    configure_logging,
    get_ini,
    Timer,
)
from config import MONGO_URL
from utils import lock
from utils import LockHeld

log = None


def update_review_queue(db):
    with Timer() as timer:
        lp = Launchpad.login_anonymously('charm-tools',
                                         'production',
                                         version='devel')

        # Entity roots for items of interests
        charm = lp.distributions['charms']
        charmers = lp.people['charmers']
        charm_contributors = lp.people['charm-contributors']

    log.debug("lp login time %0.2f", timer.duration())
    bug_statuses = [
        'New', 'Confirmed', 'Triaged', 'In Progress', 'Fix Committed']

    with Timer() as timer:
        bugs = charm.searchTasks(
            tags=['new-formula', 'new-charm'],
            tags_combinator="Any",
            status=bug_statuses)
        # Best practice  dictates that charmers be subscribed to all bugs
        # and merge proposals for official source package branches.
        charmers_bugs = charmers.searchTasks(status=bug_statuses)
        proposals = charmers.getRequestedReviews(status="Needs review")
        contributor_proposals = charm_contributors.getRequestedReviews(
            status="Needs review")

        juju = lp.projects['juju']
        doc_proposals = [mp for mp in juju.getMergeProposals()
                         if mp.target_branch_link.endswith("~juju/juju/docs")]

    log.debug("lp query time %0.2f", timer.duration())
    doc_proposals = [mp for mp in doc_proposals if not mp.queue_status in (
        'Work in progress',)]

    r_queue = list()
    for task in itertools.chain(bugs, charmers_bugs):
        entry = {
            'date_modified': task.bug.date_last_updated,
            'date_created': task.date_created,
            'summary': task.title.split('"')[1].strip(),
            'item': task.web_link,
            'status': task.status,
        }
        r_queue.append(entry)

    for proposal in itertools.chain(proposals,
                                    contributor_proposals,
                                    doc_proposals):

        parts = proposal.target_branch_link.rsplit('/', 3)
        if parts[-1] == "trunk":
            parts.pop(-1)

        target = "/".join(parts[2:])
        origin = "/".join(proposal.source_branch_link.rsplit('/', 3)[1:])

        num_comments = proposal.all_comments.total_size
        if num_comments != 0:
            last_modified = proposal.all_comments[
                num_comments - 1].date_created
        else:
            last_modified = proposal.date_created

        entry = {
            'date_modified': last_modified,
            'date_created': proposal.date_created,
            'summary': "Merge for %s from %s" % (target, origin),
            'item': proposal.web_link,
            'status': proposal.queue_status,
        }
        r_queue.append(entry)

    # Drop old collection
    log.info(
        "Dropping old review queue (%d items)", db['review_queue'].count())
    db.drop_collection('review_queue')

    # Bulk insert
    log.info("Updating review queue (%d items)" % len(r_queue))
    db['review_queue'].insert(r_queue)


def main():
    global log
    configure_logging()
    log = logging.getLogger("charm.review")
    c = pymongo.Connection(MONGO_URL, safe=True)
    db = c['juju']
    settings = get_ini()
    try:
        with lock('ingest-review',
                  int(settings['script_lease_time']) * 60, db, log):
            update_review_queue(db)
    except LockHeld, error:
        log.warn(str(error))

if __name__ == '__main__':
    main()
