# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).


from contextlib import contextmanager
from logging import getLogger
import os
import time

from bzrlib.bzrdir import BzrDir
from mock import patch


from charmworld.jobs import ingest
from charmworld.jobs.ingest import (
    add_files,
    do_bzr_update,
    get_changes,
    get_revisions,
    _rev_info,
    update_from_revisions,
)
from charmworld.models import getfs
from charmworld.testing import (
    factory,
    JobTestBase,
    TestCase,
    temp_dir,
)
from charmworld.utils import write_locked


JRH = 'J. Random Hacker <jrandom@example.com>'
YRH = 'Y. Random Hacker <yrandom@example.com>'
ZRH = 'Z. Random Hacker <zrandom@example.com>'


class TestBzrOperations(JobTestBase):

    def setUp(self):
        super(TestBzrOperations, self).setUp()
        self.log = getLogger('charm.bzr')
        self.fs = getfs(self.db)
        self.test_dir = self.use_context(temp_dir())

    def tearDown(self):
        super(TestBzrOperations, self).tearDown()

    def test_run_creates_charm_directory(self):
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        with patch('charmworld.jobs.ingest.checkout_branch'):
            with patch('charmworld.jobs.ingest.add_files'):
                do_bzr_update(charm, self.db, self.fs, self.log, self.test_dir)
        expected_path = os.path.abspath(os.path.join(
            self.test_dir, charm['series'], charm['owner'], charm['name']))
        self.assertTrue(os.path.exists(expected_path))

    def test_run_checksout_new_charms(self):
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        with patch(
                'charmworld.jobs.ingest.checkout_branch') as checkout_branch:
            with patch('charmworld.jobs.ingest.add_files'):
                do_bzr_update(charm, self.db, self.fs, self.log, self.test_dir)
        self.assertTrue(checkout_branch.called)

    def test_run_updates_existing_charms_if_not_current(self):
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        charm_path = os.path.abspath(os.path.join(
            self.test_dir,
            charm['series'],
            charm['owner'],
            charm['name'],
            charm['bname']))
        os.makedirs(charm_path)
        with patch('charmworld.jobs.ingest.branch_is_current',
                   lambda x, y, z: False):
            with patch('charmworld.jobs.ingest.add_files'):
                with patch('charmworld.jobs.ingest.update_branch') as mock:
                    do_bzr_update(
                        charm, self.db, self.fs, self.log, self.test_dir)
        mock.assert_called_with(
            self.test_dir, charm, charm_path, self.log, True)

    def test_run_does_not_update_exising_charms_if_current(self):
        ignore, charm = factory.makeCharm(self.db, branch_root=self.test_dir)
        charm_path = os.path.abspath(os.path.join(
            self.test_dir,
            charm['series'],
            charm['owner'],
            charm['name'],
            charm['bname']))
        os.makedirs(charm_path)
        with patch('charmworld.jobs.ingest.branch_is_current',
                   lambda x, y, z: True):
            with patch('charmworld.jobs.ingest.add_files'):
                with patch(
                    'charmworld.jobs.ingest.update_branch') as mock:
                    do_bzr_update(charm, self.db, self.fs, self.log,
                                  self.test_dir)
        self.assertFalse(mock.called)

    def test_add_files(self):
        with factory.charm_branch() as (branch_dir, revision):
            charm_data = {
                'branch_dir': branch_dir,
                'owner': 'charmers',
                'series': 'precise',
                'name': 'sample_charm',
                'bname': 'trunk',
                'store_data': {
                    'digest': revision,
                }
            }
            add_files(self.fs, branch_dir, charm_data, self.log)
            self.assertIn('files', charm_data)
            self.assertNotIn('icon', charm_data)

    def test_deleted_branch(self):
        # If the Launchpad branch of a charm is deleted, do_bzr_update does
        # not do anything.
        ignore, charm = factory.makeCharm(
            self.db, branch_root=self.test_dir, branch_deleted=True)
        charm_path = os.path.abspath(os.path.join(
            self.test_dir,
            charm['series'],
            charm['owner'],
            charm['name'],
            charm['bname']))
        with patch('charmworld.jobs.ingest.checkout_branch') as checkout_mock:
            with patch('charmworld.jobs.ingest.branch_is_current') as (
                    is_current_mock):
                with patch(
                    'charmworld.jobs.ingest.update_branch') as (
                        update_mock):
                    do_bzr_update(charm, self.db, self.fs, self.log,
                                    self.test_dir)
        self.assertFalse(os.path.exists(charm_path))
        self.assertFalse(checkout_mock.called)
        self.assertFalse(is_current_mock.called)
        self.assertFalse(update_mock.called)


@contextmanager
def bzr_isolation():
    with temp_dir() as temp_home:
        old_cwd = os.getcwd()
        os.chdir(temp_home)
        try:
            with patch.dict(os.environ,
                            values={
                                'HOME': temp_home,
                                'BZR_EMAIL': 'foo.bar@example.org'},
                            clear=True):
                yield temp_home
        finally:
            os.chdir(old_cwd)


class TestChangelogBzrHandling(TestCase):

    def setUp(self):
        super(TestChangelogBzrHandling)
        self.use_context(bzr_isolation())

    @staticmethod
    def num_commit(tree, num):
        return tree.commit("revision %d" % num, committer=JRH)

    def make_locked_tree(self):
        tree = BzrDir.create_standalone_workingtree('tree')
        return self.use_context(write_locked(tree))

    def test_get_revisions(self):
        tree = self.make_locked_tree()
        revisions = get_revisions(tree.branch, 10, 0)
        self.assertEqual([], revisions)
        for x in range(9):
            self.num_commit(tree, x)
        revisions = get_revisions(tree.branch, 10, 0)
        self.assertEqual(["revision %d" % num for num in range(8, -1, -1)],
                         [revision.message for revision in revisions])
        self.num_commit(tree, 9)
        revisions = get_revisions(tree.branch, 10, time.time())
        self.assertEqual(["revision %d" % num for num in range(9, -1, -1)],
                         [revision.message for revision in revisions])
        self.num_commit(tree, 10)
        revisions = get_revisions(tree.branch, 10, time.time())
        self.assertEqual(["revision %d" % num for num in range(10, 0, -1)],
                         [revision.message for revision in revisions])
        self.num_commit(tree, 11)
        revisions = get_revisions(tree.branch, 10, time.time())
        self.assertEqual(["revision %d" % num for num in range(11, 1, -1)],
                         [revision.message for revision in revisions])

    def test_get_revisions_honours_limit(self):
        tree = self.make_locked_tree()
        for x in range(15):
            self.num_commit(tree, x)
        revisions = get_revisions(tree.branch, 5, time.time())
        self.assertEqual(5, len(revisions))
        revisions = get_revisions(tree.branch, 20, time.time())
        self.assertEqual(15, len(revisions))

    def test_get_revisions_honours_since(self):
        tree = self.make_locked_tree()
        tree.commit('', timestamp=0, committer='')
        tree.commit('', timestamp=1, committer='')
        tree.commit('', timestamp=2, committer='')
        revisions = get_revisions(tree.branch, 0, 0)
        self.assertEqual(3, len(revisions))
        revisions = get_revisions(tree.branch, 0, 1)
        self.assertEqual(2, len(revisions))

    def test_limit_overrides_since(self):
        tree = self.make_locked_tree()
        tree.commit('', timestamp=0, committer='')
        tree.commit('', timestamp=1, committer='')
        tree.commit('', timestamp=2, committer='')
        revisions = get_revisions(tree.branch, since=1, limit=3)
        self.assertEqual(3, len(revisions))

    def test_rev_info(self):
        tree = self.make_locked_tree()
        revision_id = tree.commit('message1', committer=JRH)
        rev = tree.branch.repository.get_revision(revision_id)
        rev.timestamp = 34
        rev_info = _rev_info(rev, tree.branch)
        self.assertEqual({
            'authors': [JRH],
            'committer': JRH,
            'created': 34,
            'message': 'message1',
            'revno': 1,
        }, rev_info)

    @staticmethod
    def get_rev_info(branch, revision_id):
        return _rev_info(branch.repository.get_revision(revision_id), branch)

    def test_rev_info_supplied_authors(self):
        tree = self.make_locked_tree()
        revision_id = tree.commit(
            'message1', committer=JRH, authors=[YRH, ZRH])
        rev_info = self.get_rev_info(tree.branch, revision_id)
        self.assertEqual([YRH, ZRH], rev_info['authors'])
        self.assertEqual(JRH, rev_info['committer'])

    def test_get_changes(self):
        tree = self.make_locked_tree()
        first_id = self.num_commit(tree, 1)
        charm_data = get_changes('tree', 10, time.time())
        self.assertItemsEqual(['changes', 'first_change', 'last_change'],
                              charm_data.keys())
        self.assertEqual(charm_data['first_change'], charm_data['last_change'])
        self.assertEqual([charm_data['first_change']], charm_data['changes'])
        for num in range(1, 11):
            self.num_commit(tree, 1)
        charm_data = get_changes('tree', 10, time.time())
        self.assertNotEqual(charm_data['first_change'],
                            charm_data['last_change'])
        self.assertEqual(charm_data['last_change'], charm_data['changes'][0])
        self.assertNotEqual(charm_data['first_change'],
                            charm_data['changes'][-1])
        self.assertEqual(self.get_rev_info(tree.branch, first_id),
                         charm_data['first_change'])

    def test_get_changes_no_revisions(self):
        self.make_locked_tree()
        charm_data = get_changes('tree', 10, time.time())
        self.assertEqual([], charm_data['changes'])
        self.assertIs(None, charm_data['first_change'])
        self.assertIs(None, charm_data['last_change'])

    def test_branch_deleted(self):
        # update_from_revisions does not do anything if the Launchpad branch
        # of a charm is deleted.
        charm = factory.get_charm_json(branch_deleted=True)
        with patch.object(ingest, 'get_changes') as mock:
            update_from_revisions(charm)
        self.assertFalse(mock.called)
