# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from logging import getLogger
import os

import gridfs
from mock import patch

from charmworld.jobs import ingest
from charmworld.jobs.ingest import (
    scan_artifacts,
    store_provider_results,
    update_jenkins_data,
)
from charmworld.testing import RemoteURLBase


DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


def _data_file(url_frag, contents=None):
    if "/artifact/" in url_frag:
        _, fname = url_frag.rsplit("/", 1)
    else:
        fname, _ = url_frag.split("/", 1)

    if contents is None:
        with open(os.path.join(DATA_DIR, fname)) as fh:
            contents = fh.read()

    return (url_frag, contents)


class JenkinsTest(RemoteURLBase):

    url_prefix = "https://jenkins.qa.ubuntu.com/job/"

    def setUp(self):
        super(JenkinsTest, self).setUp()
        self.fs = gridfs.GridFS(self.db)
        self.log = getLogger('charm.jenkins')

    def test_jenkins(self):
        charm = {
            'branch_spec': '~charmers',
            'promulgated': True,
        }

        with patch.object(ingest, 'store_provider_results') as s:
            s.side_effect = [(1, "FAIL"), (None, None), KeyError("xz")]
            update_jenkins_data(self.db, charm, self.log)
        self.assertEqual(
            charm['tests'], {"ec2": "FAIL"})

        self.assertEqual(
            charm['test_results'], {"ec2": 1})

        charm = {
            'branch_spec': '',
            'promulgated': False,
        }
        update_jenkins_data(self.db, charm, self.log)
        self.assertEqual(
            charm,
            {
                'branch_spec': '',
                'promulgated': False,
            })

    def test_jenkins_error(self):
        # Old test data is retained if store_provider_results() fails,
        # for example, if the Jenkins server is unreachable.
        expected_test_status = {
            'ec2': 'SUCCESS',
            'openstack': 'SUCCESS',
            'local': 'FAIL',
        }
        expected_test_results = {
            'ec2': 'whatever',
            'openstack': 'whatever',
            'local': 'whatever',
        }
        charm = {
            'branch_spec': '~charmers',
            'promulgated': True,
            'tests': expected_test_status.copy(),
            'test_results': expected_test_results.copy(),
        }

        def failing_store_provider_results(*args):
            raise ValueError

        with patch.object(ingest, 'store_provider_results',
                          failing_store_provider_results):
            update_jenkins_data(self.db, charm, self.log)

        self.assertEqual(expected_test_status, charm['tests'])
        self.assertEqual(expected_test_results, charm['test_results'])

    def test_store_results(self):
        self.url_result_map = dict([
            _data_file(
                "precise-openstack-charm-mysql/lastBuild/api/json"),
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charm-graph-tests-logs.tar.gz"), "def"),
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charm-graph-tests-plans.zip"), "xyz"),
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charm-unit-tests-logs.tar.gz"), "xyz"),
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charm-revision"), "21\n\n"),
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charm-mysql.zip"), "abc"),
            # This file's existence seems to be some jenkins
            # environment cleaning issue.
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charms/precise/lodgeit/opt/lodgeit.tar.gz"), "abc")])

        with self.url_mock():
            result = store_provider_results(
                self.db,
                "openstack",
                {"name": "mysql",
                 "series": "precise",
                 "branch_spec": "~charmers/precise/mysql/trunk"},
                self.log)
        test_run = self.db.jenkins.find_one()
        self.assertEqual(test_run['revno'], 21)
        self.assertEqual(
            test_run['branch_spec'], '~charmers/precise/mysql/trunk')

        # DISABLED
        self.assertEqual(len(test_run['artifacts']), 0)
        self.assertEqual(test_run['charmrunner'], True)
        self.assertEqual(test_run['_id'], result[0])
        self.assertEqual(test_run['result'], result[1])

    def test_store_results_exists(self):
        self.url_result_map = dict([
            _data_file("precise-openstack-charm-mysql/lastBuild/api/json")])
        rid = "~charmers/precise/mysql/trunk::openstack-31"
        self.db.jenkins.insert({"_id": rid, "result": "SUCCESS"})

        with self.url_mock():
            result = store_provider_results(
                self.db,
                "openstack",
                {"name": "mysql",
                 "series": "precise",
                 "branch_spec": "~charmers/precise/mysql/trunk"},
                self.log)
        self.assertEqual(result, (rid, "SUCCESS"))

    def test_store_results_error(self):
        with self.url_mock():
            result = store_provider_results(
                self.db,
                "openstack",
                {"name": "mysql",
                 "series": "precise",
                 "branch_spec": "~charmers/precise/mysql/trunk"},
                self.log)
        self.assertEqual(result, (None, None))

    def test_revid_results(self):
        self.url_result_map = dict([
            _data_file(
                ("precise-openstack-charm-mysql/31/artifact/"
                 "charm-revision"), "21 asdf qqqq\n\n"),
        ])
        result = {'number': 31, 'artifacts': [{
            'displayPath': 'charm-revision',
            'relativePath': 'charm-revision'
        }]}
        charm = {
            "name": "mysql",
            "series": "precise",
            "branch_spec": "~charmers/precise/mysql/trunk"
        }
        with self.url_mock():
            scan_artifacts('openstack', charm, result)
        self.assertEqual(result['revno'], 21)
        self.assertEqual(result['revision-id'], 'asdf qqqq')
