# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from logging import getLogger
from requests import Response

from charmworld.charmstore import CharmStore
from charmworld.jobs.ingest import (
    addresses,
    get_address,
    update_from_store,
)
from charmworld.testing import factory
from charmworld.testing import JobTestBase


class TestStoreJob(JobTestBase):

    def setUp(self):
        super(TestStoreJob, self).setUp()
        self.log = getLogger('charm.store')

    def _mock_data(self, warning=False, error=False):
        data = {'revision': 1}
        if warning:
            data['warnings'] = 'Warning forced by mock.'
        if error:
            data['errors'] = 'Error forced by mock.'
        return data

    def falsify_session(self, store, content='[[0]]'):

        class FakeSession:

            @staticmethod
            def send(request):
                response = Response()
                response._content = content
                return response

        store.session = FakeSession

    def test_error_handling_error(self):
        handler = self.get_handler('charm.store')
        store = CharmStore()
        self.falsify_session(store)
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        address = get_address(charm, True)
        update_from_store(charm, address, self._mock_data(error=True), 'time',
                          self.log)
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("store error on %s {'errors': 'Error forced by mock.',"
                   " 'revision': 1}" % address)
        self.assertIn(err_msg, log_messages)

    def test_error_handling_warning(self):
        handler = self.get_handler(self.log.name)
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        address = get_address(charm, True)
        update_from_store(charm, address, self._mock_data(warning=True),
                          'time', self.log)
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("store error on %s {'warnings': 'Warning forced by mock.',"
                   " 'revision': 1}" % address)
        self.assertIn(err_msg, log_messages)

    def test_run(self):
        store = CharmStore()
        self.falsify_session(store, '[["2010-12-24", 4], ["2010-12-25", 1]]')
        ignore, charm = factory.makeCharm(self.db, promulgated=True)
        del charm['store_url']
        update_from_store(charm, get_address(charm, True), self._mock_data(),
                          'foo', self.log)
        data = charm['store_data']
        self.assertIn('store_checked', data.keys())
        self.assertEqual(1, data['revision'])
        self.assertEqual(
            "cs:%s/%s-%d" % (charm["series"], charm["name"], 1),
            charm['store_url'])

    def test_addresses(self):
        # The helper function addresses() returns the cs:~owner variant
        # of a store URL for unpromulgated charms.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': False,
        }
        self.assertEqual(['cs:~foo/precise/bar'], list(addresses(charm)))
        # addresses() returns the URL and the long URL for promulgated
        # charms.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': True,
        }
        self.assertEqual(
            ['cs:precise/bar', 'cs:~foo/precise/bar'],
            list(addresses(charm)))
