# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging
from pymongo import Connection
from pymongo.errors import DuplicateKeyError
from time import time

from charmworld.jobs.utils import BadDBConnection
from charmworld.jobs.utils import lock
from charmworld.jobs.utils import LockHeld
from charmworld.testing import MONGO_DATABASE
from charmworld.testing import MONGO_URL
from charmworld.testing import MongoTestBase


LOCK_NAME = u'test_lock'


class LockTest(MongoTestBase):

    def setUp(self):
        super(LockTest, self).setUp()
        self.locked_method_called = False
        self.log = logging.getLogger('locktest')
        self.log.setLevel(logging.WARN)
        self.handler = self.get_handler('locktest')

    def test_regular_locking(self):
        # If no lock with the given name is held, a lock() call
        # is successful.
        self.assertIs(None, self.db['locks'].find_one({'_id': LOCK_NAME}))
        with lock(LOCK_NAME, 1, self.db, self.log):
            # Now a lock exists.
            self.assertIsNot(
                None, self.db['locks'].find_one({'_id': LOCK_NAME}))
        # The lock has been removed.
        self.assertIs(None, self.db['locks'].find_one({'_id': LOCK_NAME}))

    def use_lock(self, exc=None, db=None):
        # Acquire a lock; optionally raise an exception.
        if db is None:
            db = self.db
        with lock(LOCK_NAME, 1, db, self.log):
            self.locked_method_called = True
            if exc is not None:
                raise exc

    def make_lock(self, lease_time, pid=1):
        return {
            '_id': LOCK_NAME,
            'lease_until': int(time() + lease_time),
            'host': 'another-host',
            'pid': pid,
        }

    def test_lock_already_held(self):
        # If another process has already created a lock, another call
        # to lock() raises a LockHeld exception.
        existing_lock = self.make_lock(5)
        self.db['locks'].insert(existing_lock)
        self.assertRaises(LockHeld, self.use_lock)

    def test_stale_lock(self):
        # Stale locks are removed.
        existing_lock = self.make_lock(-1)
        self.db['locks'].insert(existing_lock)
        self.use_lock()
        warning = self.handler.buffer[0]
        self.assertEqual(
            'Stale lock removed. Host: %(host)s, PID: %(pid)s, '
            'lease_until: %(lease_until)s' % existing_lock, warning.msg)

    def test_lock_released_when_locked_function_fails(self):
        # The lock is also released when the locked function raises an
        # exception.
        error_occurred = False
        try:
            self.use_lock(exc=ValueError)
        except ValueError:
            error_occurred = True
        self.assertTrue(error_occurred)
        self.assertIs(None, self.db['locks'].find_one({'_id': LOCK_NAME}))

    def test_duplicate_deletion_of_stale_lock(self):
        # MongoDB's Collection.remove() call does not fail when called
        # for a non-existing document. This means that two processes
        # can attempt to delete a stale lock without failing.
        locks = self.db['locks']
        lock = self.make_lock(-1)
        locks.insert(lock)
        locks.remove(lock)
        self.assertIs(None, locks.find_one(lock))
        # The second call to remove the lock does not raise an exception.
        locks.remove(lock)

    def test_lock_removal_succeeds_only_for_known_lock(self):
        # A possible race condition: Two processes detect a stale lock.
        # Process A removes the lock and creates a new lock...
        stale_lock = self.make_lock(-2, pid=1)
        lock_process_a = self.make_lock(2, pid=2)
        locks = self.db.locks
        locks.insert(lock_process_a)
        # ...before process B tries to remove the stale lock again.
        locks.remove(stale_lock)
        # The new lock still exists.
        self.assertEqual(lock_process_a, locks.find_one({'_id': LOCK_NAME}))

    def duplicate_insert(self, conn, expect_duplicate_key_error):
        # Try to insert the same document twice into the Mongo DB.
        db = conn[MONGO_DATABASE]
        locks = db['locks']
        lock_1 = self.make_lock(1, pid=42)
        locks.insert(lock_1)
        stored_lock = locks.find_one({'_id': LOCK_NAME})
        # A second insert call may succeed..
        lock_2 = self.make_lock(1, pid=43)
        if expect_duplicate_key_error:
            self.assertRaises(DuplicateKeyError, locks.insert, lock_2)
        else:
            locks.insert(lock_2)
            # ...but has no effect.
            stored_lock = locks.find_one({'_id': LOCK_NAME})
            self.assertEqual(lock_1, stored_lock)
        locks.remove({'_id': LOCK_NAME})

    def test_bad_connection_setup(self):
        # If the DB connection is created without the parameter fsync,
        # no DuplicateKeyError is raised.
        conn = Connection(MONGO_URL)
        self.duplicate_insert(conn, expect_duplicate_key_error=False)

        # A DuplicateKeyError is raised if the parameter fsync is specified.
        conn = Connection(MONGO_URL, fsync=True)
        self.duplicate_insert(conn, expect_duplicate_key_error=True)
        # Note that the value of fsync does not matter...
        conn = Connection(MONGO_URL, fsync=False)
        self.duplicate_insert(conn, expect_duplicate_key_error=True)

        # Alternatively, using safe=True ensures also that a
        # DuplicateKeyError is raised.
        conn = Connection(MONGO_URL, safe=False)
        self.duplicate_insert(conn, expect_duplicate_key_error=False)
        conn = Connection(MONGO_URL, safe=True)
        self.duplicate_insert(conn, expect_duplicate_key_error=True)

    def test_lock_creation_without_fsync_or_safe_param(self):
        # A DB connection that was created without the fsync parameter
        # or safe=True cannot be used to create a lock.
        conn = Connection(MONGO_URL)
        db = conn['locks']
        self.assertRaises(BadDBConnection, self.use_lock, db=db)
