# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

from contextlib import contextmanager
from mongoqueue import MongoQueue
from nose.tools import nottest
from pymongo.errors import DuplicateKeyError
from os import getpid
from socket import gethostname
from time import time


def get_queue(db, queue_name):
    """Get the required mongoqueues for the provided queue name.

    As some jobs only require one queue, only one queue name is mandatory.
    """
    return MongoQueue(db[queue_name], '%s-queue' % queue_name)


def parse_branch(branch_spec, series=(), commit=None):
    """Parse a branch spec into a dictionary

    A branch spec looks like the following:

      ~<owner>/charm/<series>/<charm-name>/<branch-name>

    The dictionary has the following.
      -owner
      -series
      -name
      -bname

    Returned data is augmented with commit if present
    """

    data = {}

    parts = branch_spec.split("/")

    if not len(parts) == 5:
        raise ValueError("Invalid branch spec %s" % branch_spec)

    data["owner"] = parts[0][1:]
    data["series"] = parts[2]
    data["name"] = parts[3]
    data["bname"] = parts[4]
    data["branch_spec"] = branch_spec
    data["distro_series"] = series

    if series:
        data["series"] = series[0]

    if commit:
        data["commit"] = commit

    return data

# Testing

BRANCH_TIPS = "https://api.launchpad.net/devel/charm?ws.op=getBranchTips"


class LockHeld(Exception):
    pass


class BadDBConnection(Exception):
    pass


@contextmanager
def lock(name, lease_time, db, log):
    """A simple MongoDB based mutex lock.

    :param name: The name of the mutex.
    :param lease_until: The time until which the lease is held.
    """
    locks = db['locks']
    # check that a DuplicateKeyError is raised when a doc with an existing
    # key is inserted.
    raises_key_error = False
    test_doc = {'_id': '_test'}
    try:
        locks.insert(test_doc)
        locks.insert(test_doc)
    except DuplicateKeyError:
        raises_key_error = True
    if not raises_key_error:
        raise BadDBConnection(
            'The MongoDB connection is not configured to raise '
            'DuplicateKeyErrors. Use the parameter safe=True or '
            'fsync (with any value) when creating the conenction.'
            'See test_bad_connection_setup() in '
            'charmworld/jobs/tests/test_utils.py for more details.')

    existing_lock = locks.find_one({'_id': name})
    now = time()
    if existing_lock is not None and existing_lock['lease_until'] < now:
        # Note that two calls that remove the same record do not fail, hence
        # there is no need to check the result call.
        locks.remove(existing_lock)
        log.warn(
            'Stale lock removed. Host: %(host)s, PID: %(pid)s, '
            'lease_until: %(lease_until)s' % existing_lock)
    my_lock = {
        '_id': name,
        'lease_until': now + lease_time,
        'host': gethostname(),
        'pid': getpid(),
    }
    try:
        locks.insert(my_lock)
    except DuplicateKeyError:
        existing_lock = locks.find_one({'_id': name})
        raise LockHeld(
            'Lock %s already held until %s. Current time: %s' % (
                name, existing_lock['lease_until'], now))
    try:
        yield
    finally:
        locks.remove(my_lock)


@nottest
def test():
    from charmworld.lp import get_branch_tips

    for repo, commit, series in get_branch_tips():
        print repo, parse_branch(repo)

if __name__ == '__main__':
    test()
