# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import argparse
import logging
import sys
import time

from charmworld.jobs.config import (
    CHARM_QUEUE,
    BASKET_QUEUE,
)
from charmworld.jobs.ingest import (
    run_job,
    UpdateBundleJob,
    UpdateCharmJob,
)
from charmworld.jobs.utils import get_queue
from charmworld.models import getconnection
from charmworld.models import getdb
from charmworld.search import SearchServiceNotAvailable
from charmworld.utils import (
    configure_logging,
    get_ini,
)


settings = get_ini()
INTERVAL = int(settings.get('worker_interval'))


class QueueWorker(object):
    """Processes raw charm data"""

    def __init__(self, db, queue_jobs, interval=INTERVAL):
        self.db = db
        self.queue_jobs = queue_jobs
        self.log = logging.getLogger('charm.worker')
        self.interval = interval

    def run(self, run_forever=False):
        self.log.info('Starting processing')
        return_code = 0
        for jobs in self.queue_jobs.values():
            for job in jobs:
                job.setup(db=self.db)
        while True:
            for queue, jobs in self.queue_jobs.items():
                while True:
                    item = queue.next()
                    if item is None:
                        break
                    charm_data = item.payload
                    try:
                        for job in jobs:
                            try:
                                if not run_job(
                                        job, charm_data, needs_setup=False):
                                    break
                            except SearchServiceNotAvailable as e:
                                self.log.exception(str(e))
                                queue.clear()
                                return_code = 1
                    finally:
                        item.complete()

            if not run_forever:
                self.log.info('Empty queue and not running forever; quitting.')
                break

            wait_time_in_minutes = self.interval / 60
            self.log.info(
                'Empty queue.  Waiting %s minutes to recheck.' %
                wait_time_in_minutes)
            time.sleep(self.interval)
        return return_code


def main(argv=sys.argv, db=None):
    if db is None:
        settings = get_ini()
        connection = getconnection(settings)
        db = getdb(connection, settings.get('mongo.database'))
    configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--run-forever',
        help='Keep running after the queues are processed and process more '
        'later.',
        action='store_true', dest='run_forever')
    # Explicitly pass sys.argv to make monkeypatching more reliable.
    args = parser.parse_args(argv[1:])
    queue_jobs = {
        get_queue(db, CHARM_QUEUE): [UpdateCharmJob()],
        get_queue(db, BASKET_QUEUE): [UpdateBundleJob()],
    }
    worker = QueueWorker(db, queue_jobs)
    sys.exit(worker.run(run_forever=args.run_forever))
