# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from os import environ
from os.path import join
from os.path import exists
import pymongo
from textwrap import dedent

from charmworld.models import (
    CharmSource,
    getconnection,
    getdb,
)
from charmworld.testing import (
    MongoTestBase,
    TestCase,
    temp_dir
)
from charmworld.testing.factory import random_string
from charmworld.utils import get_ini

from migrate import (
    DataStore,
    MigrationBeforeExodus,
    MissingExodusCollection,
    MissingExodusIndex,
    MultipleExodus,
    Versions,
)

INI = get_ini()
MONGO_URL = environ.get('TEST_MONGODB')
INI['mongo.url'] = MONGO_URL


SCRIPT_TEMPLATE = """

def upgrade(db, check_value=None):
    # Complete this function with the work to be done for this migration.

    return {0}

"""


def make_version_files(directory, number, runnable=True):
    """Generate a series of revisions to test against.

    :param number: The number of versions to generate from 1 - number

    """
    version_files = []
    for i in range(1, number + 1):
        filename = "{0}_{1}.py".format(str(i).zfill(3), random_string(20))
        path = join(directory, filename)

        if runnable:
            with open(path, 'w') as fh:
                # We add in the hard coded number to the template so we can
                # tell we hit the right file.
                fh.write(SCRIPT_TEMPLATE.format(i))
        else:
            # We just need the file to exist, forget about the contents.
            open(path, 'w').close()
        version_files.append(filename)
    return version_files


class TestDataStore(TestCase):
    """The data store talks to mongo."""

    def setUp(self):
        self.connection = getconnection({
            'mongo.url': MONGO_URL,
        })
        self.db = getdb(self.connection, 'juju_test')

    def tearDown(self):
        # Drop the Mongo database.
        connection = pymongo.Connection(MONGO_URL)
        connection.drop_database('juju_test')

    def test_mongo_version_starts_unset(self):
        """Unset is just None."""
        ds = DataStore(self.db)
        self.assertIsNone(ds.current_version)

    def test_mongo_version_datastore(self):
        """Versioning the db is successfull and starts at 0."""
        ds = DataStore(self.db)
        ds.version_datastore()
        self.assertEqual(0, ds.current_version)

    def test_datastore_not_reversionable(self):
        """We can't version a datastore already versioned."""
        ds = DataStore(self.db)
        ds.version_datastore()
        self.assertEqual(0, ds.current_version)

        # Now reversion and check for the exception.
        self.assertRaises(Exception, ds.version_datastore)


class TestVersions(TestCase):
    """The local versions of the files."""
    def setUp(self):
        self.tmpdir = self.use_context(temp_dir())
        self.version_files = make_version_files(self.tmpdir, 4)
        self.versions = Versions(self.tmpdir)

    def tearDown(self):
        self.ds = None
        self.version = None

    def test_latest_wo_migrations(self):
        """"""
        with temp_dir() as tmpdir:
            versions = Versions(tmpdir)
            self.assertEqual(versions.latest, 0)

    def test_latest_version(self):
        """latest returns the number of the newest revision."""
        # Note that we're saying to create up to the specified version.
        self.assertEqual(self.versions.latest, 4)

    def test_version_list(self):
        """versions is the filenames of the version files indexed by order"""
        for i, filename in enumerate(self.version_files):
            # Note that we're saying to create up to the specified version.
            self.assertEqual(filename, self.versions.versions[i + 1])

    def test_new_migration_template(self):
        """Create a new migration file and latest is updated."""
        self.versions.create_new_version_file('testing basic migration')

        self.assertEqual(5, self.versions.latest)
        self.assertEqual(
            '005_testing_basic_migration.py',
            self.versions.versions[5])
        # And this new file should exist in our versions directory
        self.assertTrue(exists(join(self.tmpdir, self.versions.versions[5])))

    def test_list_pending(self):
        self.versions.create_new_version_file('foo', content=dedent("""
            def exodus_update():
                pass
        """))
        self.versions.create_new_version_file('foo', content=dedent("""
        def update():
            pass
        """))
        self.assertEqual([
            (5, '005_foo.py', True),
            (6, '006_foo.py', False),
        ], self.versions.list_pending(4))

    def test_check_exodus(self):
        self.versions._check_exodus([])
        with self.assertRaises(MigrationBeforeExodus):
            self.versions._check_exodus([
                (1, '001_foo.py', False), (2, '002_bar.py', True)])
        self.versions._check_exodus([
            (1, '001_foo.py', True), (2, '002_bar.py', False)])
        with self.assertRaises(MultipleExodus):
            self.versions._check_exodus([
                (1, '001_foo.py', True), (2, '002_bar.py', True)])
        self.versions._check_exodus([
            (1, '001_foo.py', False), (2, '002_bar.py', False)])



class TestVersionUpgrades(TestCase):
    """Functional testing of the actual upgrade process."""
    def setUp(self):
        self.connection = getconnection({
            'mongo.url': MONGO_URL,
        })
        self.db = getdb(self.connection, 'juju_test')
        self.use_index_client(aliased=True)
        self.ds = DataStore(self.db)
        tmpdir = self.use_context(temp_dir())
        make_version_files(tmpdir, 4, runnable=True)
        self.version = Versions(tmpdir)

    def tearDown(self):
        # Drop the Mongo database.
        connection = pymongo.Connection(MONGO_URL)
        connection.drop_database('juju_test')
        self.ds = None
        self.version = None

    def test_upgrade_fails_unversioned(self):
        """upgrade will run `upgrade` methods in any migrations > current"""
        # We've not yet versioned the db, so it better blow up.
        self.assertRaises(Exception, self.version.upgrade, (self.ds, False))

    def test_upgrade_wo_migrations(self):
        """Upgrade runs silently if no migrations are created yet."""
        self.ds.version_datastore()
        tmpdir = self.use_context(temp_dir())
        versions = Versions(tmpdir)
        self.assertEqual(versions.upgrade(self.ds, self.index_client, False),
                         None)

    def test_upgrade_succeeds(self):
        """upgrade will run the methods and we get a version of 4."""
        self.ds.version_datastore()
        # We should get 4 versions done. The upgrade returns 4
        last_version = self.version.upgrade(self.ds, self.index_client, False)
        self.assertEqual(4, last_version)

        # and we can query the db for the revision of 4
        self.assertEqual(4, self.ds.current_version)

    def test_upgrade_succeeds_from_unversioned(self):
        """Forcing init no an unversioned db will upgrade successfully."""
        last_version = self.version.upgrade(self.ds, self.index_client, True)
        self.assertEqual(4, last_version)

        # and we can query the db for the revision of 4
        self.assertEqual(4, self.ds.current_version)

    def create_exodus_environment(self, index_client=None):
        if index_client is None:
            index_client = self.index_client
        versions = Versions(self.use_context(temp_dir()))
        versions.create_new_version_file(
            random_string(20), content=dedent("""
                def exodus_update(source, target, current_version):
                    pass
        """))
        source = CharmSource(self.db, index_client)
        return versions, source

    def test_upgrade_errors_if_no_exodus_db(self):
        versions, source = self.create_exodus_environment()
        aliased, = source.index_client.get_aliased()
        pending = versions.get_pending_source(source, 1)
        pending_alias = self.create_aliased(pending.index_client)
        self.addCleanup(pending_alias.delete_index)
        pending.index_client.wait_for_startup()
        with self.assertRaises(MissingExodusCollection):
            versions.upgrade(self.ds, source.index_client, True)
        self.assertEqual([aliased], source.index_client.get_aliased())

    def test_upgrade_errors_if_no_exodus_index(self):
        versions, source = self.create_exodus_environment()
        pending = versions.get_pending_source(source, 1)
        pending.create_collection()
        self.addCleanup(pending.collection.drop)
        with self.assertRaises(MissingExodusIndex):
            versions.upgrade(self.ds, self.index_client, True)
        self.assertIn(pending.collection.name,
                      pending.collection.database.collection_names())

    def create_aliased(self, index_client):
        alias_client = index_client.get_aliasable_client()
        alias_client.create_index()
        index_client.update_aliased(alias_client.index_name, [])
        return alias_client

    def create_exodus_success_environment(self):
        versions, source = self.create_exodus_environment()
        pending = versions.get_pending_source(source, 1)
        pending.create_collection()
        pending_alias = self.create_aliased(pending.index_client)
        pending_alias.put_mapping()
        return source, pending, versions

    def test_finalize_exodus_success(self):
        charm_data = {
            '_id': 'a',
            'message': 'hello',
            'owner': 'foo',
            'series': 'bar',
            'name': 'baz',
        }
        source, pending, versions = self.create_exodus_success_environment()
        aliased = source.index_client.get_aliased()
        pending.save(charm_data)
        for charm in pending._get_all('a'):
            self.assertEqual(charm_data, charm)
        versions.finalize_exodus(source, 1)
        for charm in source._get_all('a'):
            self.assertEqual(charm_data, charm)
        self.assertNotIn(pending.collection.name,
                         pending.collection.database.collection_names())
        for aliased_name in aliased:
            aliased_client = source.index_client.get_related_client(
                aliased_name)
            self.assertFalse(aliased_client.exists())
        self.assertFalse(pending.index_client.exists())

    def test_upgrade_succeeds_with_exodus(self):
        source, pending, versions = self.create_exodus_success_environment()
        pending.save({
            '_id': 'a',
            'message': 'hello',
            'owner': 'foo',
            'series': 'bar',
            'name': 'baz',
        })
        versions.upgrade(self.ds, source.index_client, True)

    def test_upgrade_errors_on_non_initial_exodus(self):
        versions = Versions(self.use_context(temp_dir()))
        versions.create_new_version_file(random_string(20))
        versions.create_new_version_file(
            random_string(20), content=dedent("""
                def exodus_update(source, target, current_version):
                    pass
        """))
        with self.assertRaises(MigrationBeforeExodus):
            versions.upgrade(self.ds, self.index_client, True)


class TestExodus(MongoTestBase):

    def setUp(self):
        super(TestExodus, self).setUp()
        self.use_index_client(aliased=True)
        self.source = CharmSource.from_request(self)
        self.versions = Versions(self.use_context(temp_dir()))

    def test_prepare_exodus_checks_validity(self):
        self.versions.create_new_version_file(
            random_string(20), content=dedent("""
                def update(source, target):
                    pass
        """))
        self.versions.create_new_version_file(
            random_string(20), content=dedent("""
                def exodus_update(source, target, current_version):
                    pass
        """))
        with self.assertRaises(MigrationBeforeExodus):
            self.versions.prepare_exodus(self.source, init=True)

    def test_prepare_exodus_initializes(self):
        with self.assertRaises(Exception) as exc_context:
            self.versions.prepare_exodus(self.source, init=False)
        self.assertEqual('Data store is not versioned',
                         str(exc_context.exception))
        self.versions.prepare_exodus(self.source, init=True)
        ds = DataStore(self.source.collection.database)
        self.assertEqual(0, ds.current_version)

    def test_prepare_exodus_acquires_pending(self):
        pending = self.versions.prepare_exodus(self.source, init=True)
        self.assertIs(None, pending)
        self.versions.create_new_version_file('foo', content=dedent("""
                def exodus_update(source, target, current_version):
                    pass
        """))
        pending = self.versions.prepare_exodus(self.source)
        self.addCleanup(pending.index_client.delete_index)
        self.assertIsInstance(pending, CharmSource)

    def test_prepare_exodus_runs_exodus_update(self):
        self.versions.create_new_version_file('foo', content=dedent("""
            def exodus_update(source, target, current_version):
                target.save({
                    '_id': 'a',
                    'message': 'hello',
                    'owner': 'foo',
                    'series': 'bar',
                    'name': 'baz'
                })
        """))
        pending = self.versions.prepare_exodus(self.source, init=True)
        self.addCleanup(pending.index_client.delete_index)
        for charm in pending._get_all('a'):
            self.assertEqual({
                    '_id': 'a',
                    'message': 'hello',
                    'owner': 'foo',
                    'series': 'bar',
                    'name': 'baz'
            }, charm)

    def test_exodus_update_runs_against_temp_but_generates_pending(self):
        self.versions.create_new_version_file('foo', content=dedent("""
            def exodus_update(source, target, current_version):
                if target.index_client.index_name == 'temp_index_pending_001':
                    raise AssertionError('Used pending name!')
                if target.collection.name == 'charms_pending_001':
                    raise AssertionError('Used pending name!')
        """))
        pending = self.versions.prepare_exodus(self.source, init=True)
        self.addCleanup(pending.index_client.delete_index)
        self.assertEqual('temp_index_pending_001',
            pending.index_client.index_name)
        self.assertEqual('charms_pending_001', pending.collection.name)

    def test_no_leftovers(self):
        db = self.source.collection.database
        index_client = self.source.index_client._client
        index_keys = index_client.status()['indices'].keys()
        index_count = len(index_client.status()['indices'])
        DataStore(db).version_datastore()
        self.source.create_collection()
        collection_names = db.collection_names()
        self.versions.create_new_version_file('foo', content=dedent("""
            def exodus_update(source, target, current_version):
                pass
        """))
        self.versions.prepare_exodus(self.source, init=True)
        self.versions.finalize_exodus(self.source, 1)
        self.assertEqual(collection_names, db.collection_names())
        self.assertEqual(index_count, len(index_client.status()['indices']))

    def test_suffix_of_source(self):
        self.versions.create_new_version_file('foo', content=dedent("""
            import re

            def exodus_update(source, target, current_version):
                target_name = target.index_client.index_name
                match = re.match('(.*)-([0-9]*)', target_name)
                if match.group(1) != source.index_client.index_name:
                    raise AssertionError('Poor name: ' + target_name)
        """))
        pending = self.versions.prepare_exodus(self.source, init=True)
        self.addCleanup(pending.index_client.delete_index)
