# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.models import (
    Bundle,
    CharmSource,
    construct_charm_id,
    FeaturedSource,
    QADataSource,
)
from charmworld.search import (
    BUNDLE,
    CHARM,
)
from charmworld.testing import (
    factory,
    MongoTestBase,
)
from charmworld.migrations import migrate
from charmworld.migrations.migrate import Versions
from charmworld.tests.test_search import put_mapping

import mock


class MigrationTestBase(MongoTestBase):

    def setUp(self):
        super(MigrationTestBase, self).setUp()
        self.versions = Versions(migrate.get_migration_path())
        self.use_context(mock.patch.object(migrate, 'configure_logging'))


class TestMigration014(MigrationTestBase):

    def get_target(self, source):
        target = self.versions.get_pending_source(source, 14)
        target.index_client.create_index()
        target.index_client.put_mapping()
        self.addCleanup(target.index_client.delete_index)
        return target

    def test_migration_updates_answers(self):
        self.use_index_client()
        charm1 = factory.get_charm_json()
        charm2 = factory.get_charm_json(name='name', series='series',
                                        owner='owner')
        charm2['qa'] = {'answer': 'data'}
        charm2['store_data']['revision'] = 5
        source = CharmSource.from_request(self)
        qa_data_source = QADataSource.from_db(self.db)
        source.save(charm1)
        source.save(charm2)
        self.assertItemsEqual([], qa_data_source.answers.find())
        target = self.get_target(source)
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        exodus_update(source, target, 11)
        self.assertEqual([{
            '_id': '~owner/series/name/5',
            'name': 'name',
            'series': 'series',
            'owner': 'owner',
            'revision': 5,
            'answers': {'answer': 'data'}
        }], list(qa_data_source.answers.find()))

    def test_migration_removes_original_qa(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        charm = factory.get_charm_json()
        charm['qa'] = {'answer': 'data'}
        source.save(charm)
        target = self.get_target(source)
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        for charm_data in source._get_all(charm['_id']):
            self.assertIn('qa', charm_data)
        exodus_update(source, target, 11)
        for charm_data in target._get_all(construct_charm_id(charm)):
            self.assertNotIn('qa', charm_data)

    def test_migration_does_nothing_if_current_version_is_12(self):
        self.use_index_client()
        charm1 = factory.get_charm_json()
        charm2 = factory.get_charm_json(name='name', series='series',
                                        owner='owner')
        charm2['qa'] = {'answer': 'data'}
        charm2['store_data']['revision'] = 5
        source = CharmSource.from_request(self)
        qa_data_source = QADataSource.from_db(self.db)
        source.save(charm1)
        source.save(charm2)
        self.assertItemsEqual([], qa_data_source.answers.find())
        target = self.get_target(source)
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        exodus_update(source, target, 12)
        self.assertEqual([], list(qa_data_source.answers.find()))
        for charm_data in source._get_all(charm2['_id']):
            self.assertIn('qa', charm_data)

    def test_migrate_featured(self):
        self.use_index_client()
        featured_source = FeaturedSource.from_db(self.db)
        charm_data = factory.get_charm_json(name='featured-charm')
        charm_data['is_featured'] = True
        source = CharmSource.from_request(self)
        source.save(charm_data)
        self.assertEqual([], list(self.db.featured.find()))
        target = self.get_target(source)
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        exodus_update(source, target, 12)
        self.assertIn(
            'featured-charm',
            [featured['name'] for featured in self.db.featured.find()])
        self.assertEqual(len(list(self.db.featured.find())), 1)

    def test_ensures_category_questions(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        target = self.get_target(source)
        self.assertItemsEqual([], self.db.qa.find())
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        exodus_update(source, target, 12)
        self.assertIn('reliable', [cat['name'] for cat in self.db.qa.find()])

    def test_migration(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        source.collection.save({
            '_id': '~foo/charms/bar/baz/trunk',
            'owner': 'foo',
            'series': 'bar',
            'name': 'baz',
            'store_data': {
                'revision': 5,
            },
        })
        target = self.get_target(source)
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        exodus_update(source, target, 12)
        self.assertEqual(['~foo/bar/baz/5'],
                         [charm['_id'] for charm in target.collection.find()])
        target_client = target.index_client
        indexed_charm_doc = target_client._client.get(
            target_client.index_name, CHARM, '~foo/bar/baz')
        # The actual document ID excludes revision, and the mongodb document
        # is under 'data'.
        self.assertEqual({
            '_id': '~foo/bar/baz',
            'data': {
                '_id': '~foo/bar/baz/5',
                'owner': 'foo',
                'series': 'bar',
                'name': 'baz',
                'store_data': {
                    'revision': 5,
                },
            }
        }, indexed_charm_doc['_source'])

    def test_removes_charm_doctype(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        old_charm_data = {
            '_id': '~foo/charms/bar/baz/trunk',
            'owner': 'foo',
            'series': 'bar',
            'name': 'baz',
            'store_data': {
                'revision': 5,
            },
            'doctype': 'charm',
        }
        charm_id = construct_charm_id(old_charm_data)
        source.collection.save(old_charm_data)
        target = self.get_target(source)
        exodus_update = self.versions.get_exodus(
            '014_id_and_search_representation.py')
        exodus_update(source, target, 12)
        for charm_data in target._get_all(charm_id):
            self.assertNotIn('doctype', charm_data)


class TestMigration015(MigrationTestBase):

    def test_removes_bundle_doctype(self):
        self.use_index_client()
        bundle, old_bundle_data = factory.makeBundle(self.db)
        old_bundle_data['doctype'] = 'bundle'
        collection = self.db.bundles
        collection.save(old_bundle_data)
        self.index_client.index_bundle(old_bundle_data)
        self.versions.run_migration(self.db, self.index_client,
                                    '015_drop_bundle_doctype.py')
        bundle_data = collection.find_one(old_bundle_data['_id'])
        self.assertNotIn('doctype', bundle_data)
        bundle_data = self.index_client.get(old_bundle_data['_id'], BUNDLE)
        self.assertNotIn('doctype', bundle_data)

    def test_ignores_doctype_free_bundle(self):
        self.use_index_client()
        bundle, old_bundle_data = factory.makeBundle(self.db)
        self.assertNotIn('doctype', old_bundle_data)
        # No exception is raised.
        self.versions.run_migration(self.db, self.index_client,
                                    '015_drop_bundle_doctype.py')


class TestMigration016(MigrationTestBase):

    def test_es_mapping_is_static(self):
        self.use_index_client(put_mapping=False)
        put_mapping(
            self.index_client,
            {'box': {'type': 'string', 'index': 'not_analyzed'}},
            dynamic=True)
        self.index_client.index_charm(factory.get_charm_json())
        old_mapping = self.index_client.get_mapping()
        self.assertIn(
            'files', old_mapping['charm']['properties']['data']['properties'])
        self.versions.run_migration(self.db, self.index_client,
                                    '016_fixed_ES_mapping_for_charms.py')
        new_mapping = self.index_client.get_mapping()
        self.assertNotIn(
            'files', new_mapping['charm']['properties']['data']['properties'])
