# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

from ConfigParser import ConfigParser
from cStringIO import StringIO
from datetime import datetime
from datetime import timedelta
from os import (
    environ,
    path,
    unlink,
)
import tempfile
from textwrap import dedent
from unittest import TestCase
import urllib

from mock import patch

from charmworld.testing import MongoTestBase
from charmworld.utils import (
    get_ini,
    get_ini_path,
)
from charmworld.utils import (
    get_session_url,
    pretty_timedelta,
    quote_key,
    quote_yaml,
    unquote_yaml,
    _process_dict_keys,
)
from charmworld.lib.config import merge_configs


class TestPrettyTimeDeltaFormatter(TestCase):

    def test_months_apart_multiple_dates(self):
        """We can pass multiple dates in to get the diff."""
        end = datetime(year=2012, month=12, day=1)
        start = datetime(year=2012, month=3, day=1)
        pretty = pretty_timedelta(start, end)

        self.assertEqual('9 months', pretty)

    def test_bad_date_order(self):
        """If we pass in the start date first we get an error."""
        start = datetime(year=2012, month=12, day=1)
        end = datetime(year=2012, month=12, day=2)
        with self.assertRaises(ValueError):
            pretty_timedelta(end, start)

    def test_days_apart(self):
        """A single delta will get us the correct days apart."""
        delta = datetime.now() - timedelta(10)
        pretty = pretty_timedelta(delta)
        self.assertEqual('10 days', pretty)


class TestIniHelper(TestCase):

    def test_get_ini_path(self):
        ini_path = get_ini_path()
        self.assertRegexpMatches(ini_path, '^/.*/.*test.ini$')
        self.assertTrue(path.exists(ini_path))
        with patch.dict(environ, clear=True):
            self.assertRegexpMatches(get_ini_path(), '^/.*/.*charmworld.ini$')

    def test_default_charmworld_ini(self):
        """By default we get a dict of our charmworld.ini file."""
        with patch.dict(environ, clear=True):
            ini = get_ini()
        self.assertEqual('Charm Browser', ini.get('project_name'))

    def test_overridden_ini(self):
        """Setting the ENV value we can override the ini read."""
        self.assertEqual('test.ini', environ['INI'])
        ini = get_ini()
        self.assertIn('project_name', ini)
        self.assertEqual('Test Charm Browser', ini.get('project_name'))
        self.assertIn('http_cache', ini)


class TestConfigSetup(TestCase):

    def makeDefaultIniFile(self, filename):
        with open(filename, 'w') as default_ini_file:
            default_ini_file.write(dedent("""
                [section_one]
                foo = one
                [section_two]
                bar = two
            """))

    def getTempFileName(self):
        result = tempfile.mktemp()
        self.addCleanup(unlink, result)
        return result

    def test_no_custom_ini_file(self):
        # If no custom .ini file exists, the contents of the default file
        # simply copied to the merged config file.
        default_ini_name = self.getTempFileName()
        result_ini_name = self.getTempFileName()
        self.makeDefaultIniFile(default_ini_name)
        merge_configs(
            result_ini_name, default_ini_name)
        result_config = ConfigParser()
        result_config.read(result_ini_name)
        self.assertEqual(
            set(('section_one', 'section_two')), set(result_config.sections()))
        self.assertEqual(['foo'], result_config.options('section_one'))
        self.assertEqual('one', result_config.get('section_one', 'foo'))
        self.assertEqual(['bar'], result_config.options('section_two'))
        self.assertEqual('two', result_config.get('section_two', 'bar'))

    def test_custom_ini_file(self):
        # If a custom .ini file exists, its settings override those from
        # the default .ini file.
        default_ini_name = self.getTempFileName()
        custom_ini_name = self.getTempFileName()
        result_ini_name = self.getTempFileName()
        with open(custom_ini_name, 'w') as custom_ini_file:
            custom_ini_file.write(dedent("""
                [section_one]
                foo = NINE
            """))
        self.makeDefaultIniFile(default_ini_name)
        merge_configs(
            result_ini_name, default_ini_name, custom_ini_name)
        result_config = ConfigParser()
        result_config.read(result_ini_name)
        self.assertEqual(
            set(('section_one', 'section_two')), set(result_config.sections()))
        self.assertEqual(['foo'], result_config.options('section_one'))
        self.assertEqual('NINE', result_config.get('section_one', 'foo'))
        self.assertEqual(['bar'], result_config.options('section_two'))
        self.assertEqual('two', result_config.get('section_two', 'bar'))

    def test_custom_ini_file_with_unexpected_data(self):
        # Definitions in the custom .ini file that do not have correspoding
        # definitions in the default .ini file are copied into the merged
        # .ini file, but a warning is printed.
        default_ini_name = self.getTempFileName()
        custom_ini_name = self.getTempFileName()
        result_ini_name = self.getTempFileName()
        with open(custom_ini_name, 'w') as custom_ini_file:
            custom_ini_file.write(dedent("""
                [section_one]
                foo = NINE
                baz = whatever
                [new_section]
                something = completely_different
            """))
        self.makeDefaultIniFile(default_ini_name)
        err_output = StringIO()
        merge_configs(
            result_ini_name, default_ini_name, custom_ini_name, err=err_output)
        result_config = ConfigParser()
        result_config.read(result_ini_name)
        self.assertEqual(
            set(('section_one', 'section_two', 'new_section')),
            set(result_config.sections()))
        self.assertEqual(['foo', 'baz'], result_config.options('section_one'))
        self.assertEqual('NINE', result_config.get('section_one', 'foo'))
        self.assertEqual('whatever', result_config.get('section_one', 'baz'))
        self.assertEqual(['bar'], result_config.options('section_two'))
        self.assertEqual('two', result_config.get('section_two', 'bar'))
        self.assertEqual(['something'], result_config.options('new_section'))
        self.assertEqual(
            'completely_different',
            result_config.get('new_section', 'something'))
        expected_warnings = (
            'Warning: %s defines sections that do not appear in %s: '
            'new_section\n'
            'Warning: Section section_one, option baz only defined in %s, '
            'not in %s\n') % (
                custom_ini_name, default_ini_name, custom_ini_name,
                default_ini_name)
        self.assertEqual(expected_warnings, err_output.getvalue())


class TestQuoteKey(MongoTestBase):

    def test_quote_key(self):
        # quote_key() replaces the characters '.', '$', '%' with escape
        # sequences.
        result = quote_key('dot . dollar $ percent %')
        self.assertEqual('dot %2E dollar %24 percent %25', result)

    def test_mongo_usage(self):
        # Key values escaped by quote_key() can be used in dictionaries
        # that are stored in a Mongo database.
        data = {
            quote_key('a.b'): 'c',
            quote_key('$d'): 'e',
            quote_key('f%'): 'g',
        }
        self.db.keytest.insert(data)
        result = self.db.keytest.find_one({quote_key('a.b'): 'c'})
        self.assertEqual(data, result)
        result = self.db.keytest.find_one({quote_key('$d'): 'e'})
        self.assertEqual(data, result)
        result = self.db.keytest.find_one({quote_key('f%'): 'g'})
        self.assertEqual(data, result)

    def test_unquote(self):
        # Strings encoded by quote_key() can be decoded by urllib.unquote().
        text = 'dot . dollar $ percent % plus +'
        self.assertEqual(text, urllib.unquote(quote_key(text)))


class TestQuoteYaml(TestCase):

    def test_simple_dict(self):
        # Dictionary keys containing the character '.', '$', '%'
        # are replaced with keys that cause no trouble when stored in
        # Mongo DB.
        data = {
            1: 'foo',
            'bar': 'baz',
            'a.b': 'c',
        }
        quote_yaml(data)
        self.assertEqual([1, 'a%2Eb', 'bar'], sorted(data))
        self.assertEqual('foo', data[1])
        self.assertEqual('baz', data['bar'])
        self.assertEqual('c', data['a%2Eb'])
        unquote_yaml(data)
        self.assertEqual([1, 'a.b', 'bar'], sorted(data))
        self.assertEqual('foo', data[1])
        self.assertEqual('baz', data['bar'])
        self.assertEqual('c', data['a.b'])

    def test_nested_dicts(self):
        inner = {
            'a': 'b',
            'c.d': 'e',
            3: 'f',
        }
        data = {
            'g': 'h',
            'i$j': inner,
        }
        quote_yaml(data)
        self.assertEqual(['g', 'i%24j'], sorted(data))
        self.assertEqual(inner, data['i%24j'])
        self.assertEqual([3, 'a', 'c%2Ed'], sorted(inner))
        self.assertEqual('e', inner['c%2Ed'])
        unquote_yaml(data)
        self.assertEqual(['g', 'i$j'], sorted(data))
        self.assertEqual(inner, data['i$j'])
        self.assertEqual([3, 'a', 'c.d'], sorted(inner))
        self.assertEqual('e', inner['c.d'])

    def test_tuple(self):
        d1 = {
            '$a': 'b',
            'c': 'd',
        }
        d2 = {
            '%e': 'f'
        }
        data = (d1, d2)
        quote_yaml(data)
        self.assertEqual(['%24a', 'c'], sorted(d1))
        self.assertEqual(['%25e'], d2.keys())
        unquote_yaml(data)
        self.assertEqual(['$a', 'c'], sorted(d1))
        self.assertEqual(['%e'], d2.keys())

    def test_key_mapper_with_conflicts(self):
        # _process_dict_keys() deals correctly with a key mapper
        # that can map a key value of a given dicitionary to another
        # existing key.
        def mapper(key):
            if key == '0':
                return '1'
            elif key == '1':
                return '0'
            else:
                return key

        data = {
            '0': 'a',
            '1': 'b',
            '2': 'c',
        }

        _process_dict_keys(data, mapper)
        self.assertEqual(data['0'], 'b')
        self.assertEqual(data['1'], 'a')
        self.assertEqual(data['2'], 'c')


class TestGetSessionURL(TestCase):

    def test_get_session_url_no_query(self):
        """Ensure URLs with no query are converted correctly."""
        settings = {
            'mongo.url': 'mongodb://example.org',
            'mongo.database': 'mydatabase',
        }
        self.assertEqual('mongodb://example.org/mydatabase.sessions',
                         get_session_url(settings))

    def test_get_session_url_query(self):
        """Ensure URLs with queries are converted correctly."""
        settings = {
            'mongo.url': 'mongodb://example.org/?replicaSet=myset',
            'mongo.database': 'mydatabase',
        }
        self.assertEqual(
            'mongodb://example.org/mydatabase.sessions?replicaSet=myset',
            get_session_url(settings))

    def test_get_session_url_multiple(self):
        """Ensure URLs with multiple hosts are converted correctly."""
        settings = {
            'mongo.url': 'mongodb://example.org,example.com/?replicaSet=myset',
            'mongo.database': 'mydatabase',
        }
        self.assertEqual(
            'mongodb://example.org,example.com/mydatabase.sessions?'
            'replicaSet=myset',
            get_session_url(settings))
