# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Helpers to generate test data for use."""

from bzrlib.bzrdir import BzrDir
from contextlib import contextmanager
from calendar import timegm
from datetime import (
    date,
    datetime,
)
import hashlib
import os
from random import randint
import random
import stat
import string
from time import time

from charmworld.charmstore import (
    get_address,
    make_store_url,
)
from charmworld.jobs.ingest import (
    JENKINS_PROVIDERS,
    process_charm,
)
from charmworld.models import (
    Bundle,
    CharmFileSet,
    construct_charm_id,
    getfs,
    options_to_storage,
)
from charmworld.utils import write_locked


def random_string(length=None):
    """Generates a random string from urandom.

    :param length: Specify the number of chars in the generated string.
    """
    chars = string.ascii_uppercase + string.digits
    str_length = length if length is not None else randint(0, 100)
    return u''.join(random.choice(chars) for x in range(str_length))


def random_url():
    """Generate a random url that is totally bogus."""
    url = "http://{0}.com".format(random_string())
    return url


def makeUser(name=None, email=None):
    """Generate a user string for changes.

    :param name: The user name; a random string if not provided.
    :param email: The user email; `name`@example.com if not provided.
    """
    if not name:
        name = random_string(10)
    if not email:
        email = '%s@example.com' % name
    return "%s <%s>" % (name, email)


def random_date(start=None, end=None):
    """Generates a random date within the range provided.

    :param start: The start of the range. The first day of the current year if
        not provided.
    :param end: The end of the range. Today if not provided.
    """
    if not start:
        start = date(date.today().year, 1, 1)
    if not end:
        end = date.today()
    increment = ((end.toordinal() - start.toordinal()) * random.random())
    return date.fromordinal(start.toordinal() + int(increment))


def random_datetime():
    return datetime.utcfromtimestamp(random.random() * 1300000000)


def makeChange(committer=None, created=None, message=None, revno=None,
               authors=None):
    """Generate a charm change.

    :param committer: The committer; a random user from `makeUser` if not
        provided.
    :param created: The date the change was committed; a random date if not
        provided.
    :param message: The change's commit message; a random string if not
        provided.
    :param revno: The change's commit's revision number; a random number if
        not provided.
    :param authors: The change's authors, as a list of name/address strings
        as returned by bzrlib.revision.Revision.get_apparent_authors()
    """
    if not committer:
        committer = makeUser()
    if not created:
        created = random_date()
    if not message:
        message = random_string(50)
    if not revno:
        revno = randint(0, 100)
    if authors is None:
        authors = ['Foo <foo@example.com>', 'Bar <bar@example.com>']

    if isinstance(created, date):
        # Change `created` dates should be timestamps.
        created = timegm(created.timetuple())
    return {
        'authors': authors,
        'committer': committer,
        'created': created,
        'message': message,
        'revno': revno}


def get_payload_json(name=None, bname=None, series=None, owner=None,
                     promulgated=False, branch_deleted=False):
    if not series:
        series = 'precise'
    if not bname:
        bname = 'trunk'
    if not owner:
        owner = "charmers"
    if not name:
        name = random_string(55)
    return {
        "bname": bname,
        "branch_deleted": branch_deleted,
        "branch_spec": "~%s/charms/%s/%s/trunk" % (owner, series, name),
        "name": name,
        "owner": owner,
        "promulgated": promulgated,
        "series": series,
        "commit": "jdoe@example.com-20120723223205-n5zke9r9ahb40sue",
    }


def add_store_data(charm, promulgated=False, charm_error=False, revision=1):
    address = get_address(charm, promulgated)
    charm["store_url"] = make_store_url(1, address)
    store_data = {
        "store_checked": "Thu Dec  6 17:42:05 2012"
    }
    if charm_error:
        store_data.update({
            'errors': ['entry not found'],
            'revision': 0,
        })
    else:
        store_data.update({
            "digest": "jdoe@example.com-20120723223205-n5zke9r9ahb40su",
            "revision": revision,
            "sha256": "107c230dc3481f8b8de9decfff6439a23caae2ea670bddb413ddf1",
        })
    charm['store_data'] = store_data


def get_charm_json(name=None, changes=None, subordinate=False, tests=None,
                   with_real_files=False, branch_root=None, bname=None,
                   series=None, owner=None, description=None, revision=1,
                   summary=None, revno=12, maintainer=None,
                   commit_message='maintainer', provides=None, requires=None,
                   options=None, files=None, charm_error=False,
                   date_created=None, downloads=0,
                   downloads_in_past_30_days=0,
                   promulgated=False, categories=None, branch_deleted=False,
                   proof=None, payload=None, empty_branch=False, hash_=None,
                   store_revision=1):
    """Return the json of a charm."""
    if not description:
        description = """byobu-class provides remote terminal access through
            a web browser,\nusing https, ajaxterm, and byobu.  It operates in a
            one-writer,\nmany reader mode.\n"""
    if not changes:
        changes = [makeChange()]
    if date_created is None:
        date_created = random_datetime()
    if branch_root is None:
        if with_real_files:
            branch_root = os.path.split(__file__)[0]
            branch_root = os.path.join(branch_root, 'data')
            name = 'sample_charm'
        else:
            branch_root = '/var/charms/precise/charmers'
    if maintainer is None:
        maintainer = "John Doe <jdoe@example.com>"
    if summary is None:
        summary = "Remote terminal classroom over https"
    if provides is None:
        provides = {
            "website": {
                "interface": "https",
            }
        }
    if requires is None:
        requires = {
            'database': {
                'interface': 'mongodb',
            }
        }
    if options is None:
        options = {
            'script-interval': {
                'type': 'int',
                'default': 5,
                'description': 'The interval between script runs.'
            },
        }
    options = options_to_storage(options)
    if files is None:
        files = {
            'readme': {
                'filename': 'README',
                'subdir': '',
            },
            'install': {
                'filename': 'install',
                'subdir': 'hooks',
            },
        }
    if proof is None:
        proof = {
            "w": [
                " no README file",
                " (install:31) - hook accesses EC2 metadata service directly"
            ]
        }

    if categories is None:
        categories = []
    elif not isinstance(categories, list):
        raise ValueError('Categories must be a list.')
    if payload is None:
        charm = get_payload_json(name, bname, series, owner, promulgated,
                                 branch_deleted)
    else:
        for key in ['name', 'bname', 'series', 'owner']:
            if locals()[key] is not None:
                raise ValueError(
                    'Cannot specify "%s" if payload supplied.' % key)
        charm = dict(payload)
    add_store_data(charm, promulgated, charm_error, store_revision)
    charm.update({
        '_id': construct_charm_id(charm),
        'address': get_address(charm, promulgated),
        "branch_dir": os.path.join(branch_root, charm['name']),
        "categories": categories,
        "changes": changes,
        "description": description,
        "hooks": [
            "install",
            "start",
            "stop",
            "website-relation-joined"
        ],
        "maintainer": maintainer,
        "proof": proof,
        "provides": provides,
        'requires': requires,
        "revision": revision,
        "summary": summary,
        'config': {
            'options': options,
        },
        'files': files,
        'downloads': downloads,
        'downloads_in_past_30_days': downloads_in_past_30_days,
        'date_created': date_created.isoformat(),
    })
    if empty_branch:
        charm["first_change"] = None
        charm["last_change"] = None
    else:
        charm["first_change"] = {
            "committer": "Jane Doe <jane@sample.com>",
            "created": 1308093138.892,
            "message": "initial checkin\n",
            "revno": 1
        }
        charm["last_change"] = {
            "committer": "John Doe <jdoe@example.com>",
            "created": 1343082725.06,
            "message": commit_message,
            "revno": revno,
        }
    if tests is not None:
        charm['tests'] = tests
        charm['test_results'] = {}
        for provider, status in tests.items():
            charm['test_results'][provider] = makeTestResult(
                charm['name'], charm['branch_spec'], status, provider)
    if subordinate:
        charm['subordinate'] = True
    if hash_ is not None:
        charm['hash'] = hash_
    return process_charm(charm)


def makeCharm(db, *args, **kwargs):
    charm = get_charm_json(*args, **kwargs)
    record_id = db.charms.insert(charm)
    return record_id, charm


def get_bundle_data(name=None, owner=None, basket=None, series='precise',
                    title='', description='', services=dict(),
                    relations=dict(), promulgated=False, branch_deleted=False,
                    ):
    if name is None:
        name = random_string(10)
    if owner is None:
        owner = random_string(10)
    if basket is None:
        basket = "%s/%d" % (random_string(10), randint(0, 100))
    data = dict(series=series,
                relations=relations,
                services=services)
    return dict(basket=basket,
                branch_deleted=branch_deleted,
                data=data,
                description=description,
                name=name,
                owner=owner,
                promulgated=promulgated,
                title=title)


def makeBundle(db, *args, **kwargs):
    bundle_data = get_bundle_data(*args, **kwargs)
    bundle = Bundle(bundle_data)
    bundle_data['_id'] = bundle.id
    db.bundles.insert(bundle_data)
    return bundle, bundle_data


def make_charm_file(db, charm, path, content=None):
    fs = getfs(db)
    charm_file = CharmFileSet.make_charm_file(
        fs, charm, path, os.path.basename(path))
    if content is None:
        content = random_string(10).encode('utf-8')
    charm_file.save(content)
    return charm_file


def makeTestResult(db, name, branch_spec, status, provider,
                   branch_revision=None, timestamp=None, test_number=None):
    """Create a test record.
    """
    if provider is None:
        provider = random.choice(JENKINS_PROVIDERS)
    if test_number is None:
        test_number = randint(20, 40)
    if timestamp is None:
        timestamp = int(time()) * 1000
    test_id = '%s::%s-%s' % (branch_spec, provider, test_number)
    test_result = {
        '_id': test_id,
        'branch_spec': branch_spec,
        'number': test_number,
        'provider': provider,
        'result': status,
        'timestamp': timestamp,
        'url': 'https://jenkins.qa.ubuntu.com/job/precise-%s-charm-%s/%s' % (
            provider, name, test_number)
    }
    if branch_revision is None:
        branch_revision = randint(20, 40)
        test_result['revno'] = branch_revision
    db.jenkins.insert(test_result)
    return test_id


def makeSSOResponse(sso_type='login', groups=None):
    if sso_type == 'login':
        payload = {
            u'openid.op_endpoint': u'https://login.ubuntu.com/+openid',
            u'openid.sig': u'o3OoughdEky+xKkryB2USvcwU2w=',
            u'openid.sreg.language': u'en',
            u'openid.return_to': u'http://127.0.0.1/auth_callback?janrain_no',
            u'openid.ns': u'http://specs.openid.net/auth/2.0',
            u'janrain_nonce': u'2012-12-17T18:35:57ZDvqeup',
            u'openid.response_nonce': u'2012-12-17T18:36:42Z3Xqn5M',
            u'openid.ns.sreg': u'http://openid.net/extensions/sreg/1.1',
            u'openid.sreg.fullname': u'Sample Person',
            u'openid.claimed_id': u'https://login.ubuntu.com/+id/MSbTCYM',
            u'openid.mode': u'id_res',
            u'openid.sreg.nickname': u'sample',
            u'openid.sreg.timezone': u'America/Detroit',
            u'openid.signed': u"""assoc_handle,
               claimed_id,
               identity,
               mode,
               ns,
               ns.sreg,
               op_endpoint,
               response_nonce,
               return_to,
               signed,
               sreg.email,
               sreg.fullname,
               sreg.language,
               sreg.nickname,
               sreg.timezone""",
            u'openid.assoc_handle': u'{HMAC-SHA1}{50cf6493}{6QI6Fw==}',
            u'openid.identity': u'https://login.ubuntu.com/+id/MSbTCYM',
            u'openid.sreg.email': u'sample@example.com',
            u'openid.lp.is_member': '',
        }

        if groups:
            payload['openid.lp.is_member'] = ','.join(groups)
        return payload

    elif sso_type == 'cancel':
        return {
            u'janrain_nonce': u'2012-12-17T18:29:39ZKnPAWl',
            u'openid.mode': u'cancel',
            u'openid.ns': u'http://specs.openid.net/auth/2.0',
        }


def makeQAAssessment():
    """Data stored in the qa of a Charm based off the forms/qa_assessment"""
    from collections import OrderedDict
    return OrderedDict([
        ('work_needed_link', u'http://google.com'),
        ('notes', u'lots of notes'),
        ('__start__', u'Reliable:mapping'),
        ('__start__', u'4847e034bb0a55fcbc8a3380d6a3ab80:rename'),
        ('deformField15', u'1'),
        ('__end__', u''),
        ('__start__', u'a60b40bce263c47559d32435a89756fe:rename'),
        ('deformField16', u'1'),
        ('__end__', u''),
        ('__start__', u'7ff6cc997b141399bcdf2c338fb4031c:rename'),
        ('deformField17', u'1'),
        ('__end__', u''),
        ('__start__', u'9a736d9ab5ca1e16ddd942e7a62b8820:rename'),
        ('deformField18', u'0'),
        ('__end__', u''),
        ('__start__', u'51dceb2909262166283ac17f2af504e7:rename'),
        ('deformField19', u'0'),
        ('__end__', u''),
        ('__start__', u'9659919b570c6ef9a25df7261375e6bb:rename'),
        ('deformField20', u'0'),
        ('__end__', u''),
        ('__start__', u'e3c804e2ac59ebb8cb73ac26d43cb260:rename'),
        ('deformField21', u'0'),
        ('__end__', u''),
        ('__start__', u'4cb0090569d839a2bffee6fba06ac2eb:rename'),
        ('__end__', u''),
        ('__start__', u'34e37456112085f7b71b6347f8184a73:rename'),
        ('__end__', u''),
        ('__end__', u'Reliable:mapping'),
        ('__start__', u'Secure:mapping'),
        ('__start__', u'ffa1dc58332d3a0e7a957e2a55007b4f:rename'),
        ('deformField25', u'0'),
        ('__end__', u''),
        ('__start__', u'190e6b028777b602190de1763660ae6f:rename'),
        ('deformField26', u'1'),
        ('__end__', u''),
        ('__start__', u'83d5c7dc3ca94ff5a966099c7c722e08:rename'),
        ('deformField27', u'1'),
        ('__end__', u''),
        ('__start__', u'39e130a521cf04bd9af11138dcdf8d8c:rename'),
        ('deformField28', u'0'),
        ('__end__', u''),
        ('__end__', u'Secure:mapping'),
        ('__start__', u'Flexible:mapping'),
        ('__start__', u'3a57079d741ccc48e22db219b0aec68e:rename'),
        ('deformField30', u'0'),
        ('__end__', u''),
        ('__start__', u'c913e2509bcbf1cc027e95dae1f8ba7f:rename'),
        ('deformField31', u'1'),
        ('__end__', u''),
        ('__end__', u'Flexible:mapping'),
        ('__start__', u'Handle Data:mapping'),
        ('__start__', u'69be998e2aee961adfb12773c6796957:rename'),
        ('deformField33', u'0'),
        ('__end__', u''),
        ('__start__', u'afc32c74c6b99b784b41af5e44d43981:rename'),
        ('deformField34', u''),
        ('__end__', u''),
        ('__start__', u'afc32c74c6b99b784b41af5e44d43981:rename'),
        ('deformField35', u''),
        ('__end__', u''),
        ('__end__', u'Handle Data:mapping'),
        ('__start__', u'Scaleable:mapping'),
        ('__start__', u'49eb3b0d38a795f41c59d62d46a3b887:rename'),
        ('deformField37', u'0'),
        ('__end__', u''),
        ('__start__', u'd048cfb0691d53c6eb3aced71340cab3:rename'),
        ('deformField38', u'1'),
        ('__end__', u''),
        ('__start__', u'f081c5d19281a7decfd06bf754c82386:rename'),
        ('deformField39', u'1'),
        ('__end__', u''),
        ('__start__', u'191a39c1b62a1ab8c97af36eb23e4af8:rename'),
        ('deformField40', u'1'),
        ('__end__', u''),
        ('__start__', u'b10cadcaecfc455cf5b5c4521edb7d86:rename'),
        ('deformField41', u'0'),
        ('__end__', u''),
        ('__end__', u'Scaleable:mapping'),
        ('__start__', u'Easy to Deploy:mapping'),
        ('__start__', u'899c01c6ba4103348c71b968d9c6b91b:rename'),
        ('deformField43', u'0'),
        ('__end__', u''),
        ('__start__', u'98331d802f2ff09e4dd256d0b488c20d:rename'),
        ('deformField44', u'0'),
        ('__end__', u''),
        ('__start__', u'419f40223e91f5bb1f6a734b800d2f1d:rename'),
        ('deformField45', u'0'),
        ('__end__', u''),
        ('__start__', u'9bcf839c9b5b85974b5a68b65e1784d3:rename'),
        ('deformField46', u'1'),
        ('__end__', u''),
        ('__start__', u'1fc3e7fe66dca5d512a54ee178336f78:rename'),
        ('deformField47', u'1'),
        ('__end__', u''),
        ('__start__', u'77dd5c463f0c4bc455633307c87957cf:rename'),
        ('deformField48', u'1'),
        ('__end__', u''),
        ('__start__', u'64a0c6cd62f82f9a2a2f2971e8732f39:rename'),
        ('deformField49', u'0'),
        ('__end__', u''),
        ('__end__', u'Easy to Deploy:mapping'),
        ('__start__', u'Responsive to DevOps Needs:mapping'),
        ('__start__', u'6d74092ba1b34019f7276141820fcac6:rename'),
        ('deformField51', u'1'),
        ('__end__', u''),
        ('__start__', u'348be0e0e93d66d3084c8709dffc260e:rename'),
        ('deformField52', u'0'),
        ('__end__', u''),
        ('__start__', u'76634ca7dc25e6071ff327b78209fbc2:rename'),
        ('deformField53', u'0'),
        ('__end__', u''),
        ('__start__', u'9293ffcaf0cca86c793bd92faf21360f:rename'),
        ('deformField54', u'1'),
        ('__end__', u''),
        ('__end__', u'Responsive to DevOps Needs:mapping'),
        ('__start__', u'Upstream Friendly:mapping'),
        ('__start__', u'c0b6382136022a6fdc6c20a6ffc45ca5:rename'),
        ('deformField56', u'0'),
        ('__end__', u''),
        ('__start__', u'5fd70986787dae347b7a579f8c25fcd2:rename'),
        ('deformField57', u'1'),
        ('__end__', u''),
        ('__start__', u'375baad73faa59703208fccd84640ded:rename'),
        ('deformField58', u'1'),
        ('__end__', u''),
        ('__start__', u'ca575c0c1fce388c18b1cd13fa14e293:rename'),
        ('deformField59', u'1'),
        ('__end__', u''),
        ('__end__', u'Upstream Friendly:mapping'),
    ])


def makeQAScore():
    return {
        "data_handling": {
            "data_handling_2": "1",
            "data_handling_0": "1",
            "data_handling_1": "0"
        },
        "secure": {
            "secure_2": "0",
            "secure_3": "0",
            "secure_0": "1",
            "secure_1": "1"
        },
        "notes": "More work",
        "scalable": {
            "scalable_2": "0",
            "scalable_3": "0",
            "scalable_0": "1",
            "scalable_1": "1",
            "scalable_4": "0"
        },
        "responsive": {
            "responsive_3": "0",
            "responsive_2": "1",
            "responsive_1": "1",
            "responsive_0": "1"
        },
        "upstream": {
            "upstream_0": "1",
            "upstream_1": "1",
            "upstream_2": "0",
            "upstream_3": "0"
        },
        "work_needed_link": "Need work",
        "flexible": {
            "flexible_0": "1",
            "flexible_1": "0"
        },
        "easy_deploy": {
            "easy_deploy_0": "1",
            "easy_deploy_1": "1",
            "easy_deploy_2": "1",
            "easy_deploy_3": "0",
            "easy_deploy_4": "0",
            "easy_deploy_5": "0",
            "easy_deploy_6": "0"
        },
        "reliable": {
            "reliable_1": "1",
            "reliable_0": "1",
            "reliable_3": "0",
            "reliable_2": "0",
            "reliable_5": "1",
            "reliable_4": "",
            "reliable_7": "0",
            "reliable_6": "1",
            "reliable_8": "0"
        }
    }


def makeQAQuestionGroup():
    """Generate the dict needed for the qa collection."""

    category = (
        u'reliable',
        u'Reliable',
        [
            (u'AWS', 1, u''),
            (u'HP Cloud', 1, u''),
            (u'OpenStack', 1, u''),
            (u'LXC', 1, u''),
            (u'MAAS', 1, u''),
            (u'Check for integrity from upstream source', 1, u''),
            (u'Fail gracefully if upstream source goes missing', 1, u''),
            (u'Contain a suite of tests with the charm that pass', 1, u''),
            (u'Passes tests from Jenkins on jujucharms.com', 1, u''),
        ]
    )

    category_dict = {
        'name': category[0],
        'description': category[1],
        'questions': [{
            'id': hashlib.md5(q[0]).hexdigest(),
            'description': q[0],
            'points': q[1],
            'extended_description': q[2]
        } for q in category[2]]
    }

    return category_dict


def make_review_entry(**kwargs):
    """Create an entry for the review queue.

    keyword arguments override the defaults.
    """
    entry = {
        'date_modified': random_datetime(),
        'date_created': random_datetime(),
        'summary': random_string(10),
        'item': random_string(10),
        'status': 'Needs review',
    }
    entry.update(kwargs)
    return entry


@contextmanager
def bzr_branch():
    """Create a Bazaar branch."""
    from charmworld.jobs.tests.test_bzr import bzr_isolation
    with bzr_isolation() as cwd:
        branch_dir = os.path.join(cwd, 'branch/trunk')
        os.makedirs(branch_dir)
        tree = BzrDir.create_standalone_workingtree(branch_dir)
        yield cwd, branch_dir, tree


@contextmanager
def charm_branch(charm_path=None, extra=[]):
    """Create a Bazaar branch from the files in charm_path."""
    # Avoid circular imports.
    if charm_path is None:
        charm_path = os.path.split(__file__)[0]
        charm_path = os.path.join(charm_path, 'data/sample_charm')
    with bzr_branch() as (cwd, branch_dir, tree):
        with write_locked(tree):
            for dirpath, dirnames, filenames in os.walk(charm_path):
                src_paths = [os.path.join(dirpath, name) for name in dirnames]
                rel_paths = [
                    os.path.relpath(path, charm_path) for path in src_paths]
                dst_paths = [
                    os.path.join(branch_dir, path) for path in rel_paths]
                for src, dst in zip(src_paths, dst_paths):
                    if stat.S_ISLNK(os.lstat(src).st_mode):
                        os.symlink(os.readlink(src), dst)
                    else:
                        os.mkdir(dst)
                tree.add(rel_paths)
                src_paths = [os.path.join(dirpath, name) for name in filenames]
                rel_paths = [
                    os.path.relpath(path, charm_path) for path in src_paths]
                dst_paths = [
                    os.path.join(branch_dir, path) for path in rel_paths]
                for src, dst in zip(src_paths, dst_paths):
                    if stat.S_ISLNK(os.lstat(src).st_mode):
                        os.symlink(os.readlink(src), dst)
                    else:
                        with open(dst, 'w') as out:
                            out.write(open(src).read())
                tree.add(rel_paths)

            revision = tree.commit('commit 1')
            for func in extra:
                result = func(cwd, tree)
                if result is not None:
                    revision = result
        yield branch_dir, revision
