
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import (
    date,
    datetime,
)
import json
import urlparse

from requests import Response

from charmworld.charmstore import (
    CharmStore,
    get_address,
    get_store_charm_info,
)
from charmworld.testing import (
    factory,
    TestCase,
)


class FakeSessionBase:

    @classmethod
    def send(cls, request):
        parts = urlparse.urlsplit(request.url)
        query = urlparse.parse_qs(parts.query)
        if request.body is not None:
            data = urlparse.parse_qs(request.body)
        else:
            data = None
        response = Response()
        response._content = cls.parsed_send(request, parts, query, data)
        return response


class TestCharmStore(TestCase):

    def assertStoreURL(self, expected_path, url):
        self.assertEqual(CharmStore.STAT_URL + expected_path, url)

    def test_get_stats_range_request(self):
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        start = date(2011, 01, 02)
        end = date(2012, 12, 31)
        cs = CharmStore()
        req = cs.get_stats_range_request(charm, start, end)
        self.assertEqual('GET', req.method)
        self.assertStoreURL('/charm-bundle:foo:bar:baz', req.url)
        self.assertEqual({
            'start': '2011-01-02',
            'end': '2012-12-31',
            'format': 'json',
            'by': 'day',
        }, req.params)
        startdatetime = datetime(2011, 02, 03)
        req = cs.get_stats_range_request(charm, startdatetime, startdatetime)
        self.assertEqual('2011-02-03', req.params['start'])
        self.assertEqual('2011-02-03', req.params['end'])
        charm['store_url'] = 'cs:foo/bar'
        url = cs.get_stats_range_request(charm, start, end).url
        self.assertEqual(
            'https://store.juju.ubuntu.com/stats/counter/charm-bundle:foo:bar',
            url)

    def test_get_stats_range_request_wo_dates(self):
        """The start and end dates are optional

        It will fetch total counts without the start and end dates.

        """
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        cs = CharmStore()
        req = cs.get_stats_range_request(charm)
        self.assertEqual('GET', req.method)
        self.assertStoreURL('/charm-bundle:foo:bar:baz', req.url)
        self.assertEqual({
            'format': 'json'
        }, req.params)

    def test_get_download_counts(self):
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        start = date(2011, 01, 02)
        end = date(2012, 12, 31)
        cs = CharmStore()

        class FakeSession(FakeSessionBase):

            @staticmethod
            def parsed_send(request, parts, query, data):
                self.assertEqual('/stats/counter/charm-bundle:foo:bar:baz',
                                 parts.path)
                self.assertEqual({
                    'start': ['2011-01-02'],
                    'end': ['2012-12-31'],
                    'by': ['day'],
                    'format': ['json']}, query)
                return '[[0]]'

        cs.session = FakeSession()
        result = cs.get_download_counts(charm, start, end)
        self.assertEqual([], result)

        class FakeSession2:
            @staticmethod
            def send(request):
                response = Response()
                response._content = '[["2013-15-34", 10]]'
                return response

        cs.session = FakeSession2()
        result = cs.get_download_counts(charm, start, end)
        self.assertEqual([["2013-15-34", 10]], result)

    def test_count_downloads_in_days(self):
        charm = factory.get_charm_json(series='foo', name='bar', owner='baz')
        today = date(2011, 01, 02)
        cs = CharmStore()

        class FakeSession:

            @staticmethod
            def send(request):
                parts = urlparse.urlsplit(request.url)
                self.assertEqual('/stats/counter/charm-bundle:foo:bar:baz',
                                 parts.path)
                query = urlparse.parse_qs(parts.query)
                self.assertEqual(['2011-01-02'], query['end'])
                self.assertEqual(['2010-12-03'], query['start'])
                response = Response()
                response._content = '[["2010-12-24", 4], ["2010-12-25", 1]]'
                return response

        cs.session = FakeSession()

        self.assertEqual(5, cs.count_downloads_in_days(charm, 30, today))

    def test_get_charms_info_request(self):
        req = CharmStore.get_charms_info_request(
            ['cs:asdf/qsf', 'cs:~foo/bar/baz'])
        self.assertEqual('POST', req.method)
        self.assertEqual(CharmStore.STORE_URL + '/charm-info', req.url)
        self.assertEqual('0', req.data['stats'])
        self.assertEqual(['cs:asdf/qsf', 'cs:~foo/bar/baz'],
                         req.data['charms'])

    def test_get_charms_info(self):
        class FakeSession(FakeSessionBase):
            @staticmethod
            def parsed_send(req, parts, query, data):
                self.assertEqual(
                    data['charms'], ['cs:asdf/qsf', 'cs:~foo/bar/baz'])
                return json.dumps({})
        store = CharmStore()
        store.session = FakeSession
        info = store.get_charms_info(['cs:asdf/qsf', 'cs:~foo/bar/baz'])
        self.assertEqual({}, info)


MISSING_CHARM = {'revision': 0, 'errors': ['entry not found']}


class FakeCharmStore:

    def __init__(self, store_data=None):
        if store_data is None:
            store_data = {}
        self.store_data = store_data

    def get_charms_info(self, addresses):
        return dict((a, self.store_data.get(a, MISSING_CHARM))
                    for a in addresses)


class TestGetStoreCharmInfo(TestCase):

    def test_get_store_charm_info(self):
        payload1 = factory.get_payload_json()
        addr1 = get_address(payload1, False)
        payload2 = factory.get_payload_json()
        addr2 = get_address(payload2, False)
        payload3 = factory.get_payload_json()
        addr3 = get_address(payload3, False)
        store = FakeCharmStore({addr1: {'1'}, addr2: {'2'}})
        result = get_store_charm_info(store, [payload1, payload2, payload3])
        self.assertEqual(
            [(addr1, {'1'}), (addr2, {'2'}), (addr3, MISSING_CHARM)], result)
        result = get_store_charm_info(store, [payload3, payload1, payload2])
        self.assertEqual(
            [(addr3, MISSING_CHARM), (addr1, {'1'}), (addr2, {'2'})], result)

    def test_short_address(self):
        payload = factory.get_payload_json(promulgated=True)
        short_addr = get_address(payload, True)
        store = FakeCharmStore({short_addr: {'foo': 'bar'}})
        result = get_store_charm_info(store, [payload])
        self.assertEqual([(short_addr, {'foo': 'bar'})], result)

    def test_long_address_if_not_promulgated(self):
        payload = factory.get_payload_json(promulgated=False)
        short_addr = get_address(payload, True)
        long_addr = get_address(payload, False)
        store = FakeCharmStore(
            {short_addr: {'foo': 'bar'}, long_addr: {'foo': 'bar'}})
        result = get_store_charm_info(store, [payload])
        self.assertEqual([(long_addr, {'foo': 'bar'})], result)

    def test_long_address_if_short_not_in_store(self):
        payload = factory.get_payload_json(promulgated=True)
        long_addr = get_address(payload, False)
        store = FakeCharmStore(
            {long_addr: {'foo': 'bar'}})
        result = get_store_charm_info(store, [payload])
        self.assertEqual([(long_addr, {'foo': 'bar'})], result)

    def test_long_address_if_neither_in_store(self):
        payload = factory.get_payload_json(promulgated=True)
        long_addr = get_address(payload, False)
        store = FakeCharmStore({})
        result = get_store_charm_info(store, [payload])
        self.assertEqual([(long_addr, MISSING_CHARM)], result)
