# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mock import patch

from charmworld.health import (
    check_api2,
    check_api3,
    check_elasticsearch,
    check_mongodb,
)
from charmworld.models import FeaturedSource
from charmworld.testing import (
    factory,
    ViewTestBase,
)


class TestCheckMixin:

    def assertCheck(self, check, name, status, remark=None):
        self.assertEqual(name, check.name)
        self.assertEqual(status, check.status)
        if remark is not None:
            self.assertEqual(remark, check.remark)

    @staticmethod
    def makeFailCallable():
        def failing(*args, **kwargs):
            raise ValueError
        return failing


class HealthTestCase(ViewTestBase, TestCheckMixin):

    def test_check_mongodb_pass(self):
        factory.makeCharm(self.db)
        check = check_mongodb(self.getRequest())
        remark = '1 charms found'
        self.assertCheck(check, 'charms collection', 'Pass', remark)

    def test_check_mongodb_missing_data_fail(self):
        check = check_mongodb(self.getRequest())
        remark = 'There are no charms. Is ingest running?'
        self.assertCheck(check, 'charms collection', 'Fail', remark)

    def test_check_mongodb_no_collection_fail(self):
        request = self.getRequest()
        with patch.object(request, 'db', new_callable=self.makeFailCallable):
            check = check_mongodb(request)
        self.assertCheck(check, 'charms collection', 'Fail')

    def test_check_elasticsearch_pass(self):
        charm_data = factory.makeCharm(self.db)[1]
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        check = check_elasticsearch(self.getRequest())
        remark = '1 matches found'
        self.assertCheck(check, 'charms index', 'Pass', remark)

    def test_check_elasticsearch_missing_data_fail(self):
        self.use_index_client()
        request = self.getRequest()
        check = check_elasticsearch(request)
        remark = 'There are no matches. Is ingest running?'
        self.assertCheck(check, 'charms index', 'Fail', remark)

    def test_check_elasticsearch_no_index_fail(self):
        request = self.getRequest()
        # The index is not setup.
        check = check_elasticsearch(request)
        self.assertCheck(check, 'charms index', 'Fail')

    def _check_api_pass(self, function, version):
        charm_data = factory.makeCharm(self.db, downloads=5)[1]
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        FeaturedSource.from_db(self.db).set_featured(charm_data, 'charm')
        check = function(self.getRequest())
        remark = 'Interesting has new, popular, and featured.'
        self.assertCheck(check, 'API%s interesting' % version, 'Pass', remark)

    def _check_api_missing_data_fail(self, function, version):
        charm_data = factory.makeCharm(self.db)[1]
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        # There are no featured or downloaded charms.
        check = function(self.getRequest())
        remark = 'Interesting is missing new, popular, or featured.'
        self.assertCheck(check, 'API%s interesting' % version, 'Fail', remark)

    def _check_api_no_matches_fail(self, function, version):
        request = self.getRequest()
        # The index is not setup for the interesting search.
        check = function(request)
        self.assertCheck(check, 'API%s interesting' % version, 'Fail')

    def test_check_api2_pass(self):
        self._check_api_pass(check_api2, 2)

    def test_check_api2_missing_data_fail(self):
        self._check_api_missing_data_fail(check_api2, 2)

    def test_check_api2_no_matches_fail(self):
        self._check_api_no_matches_fail(check_api2, 2)

    def test_check_api3_pass(self):
        self._check_api_pass(check_api3, 3)

    def test_check_api3_missing_data_fail(self):
        self._check_api_missing_data_fail(check_api3, 3)

    def test_check_api3_no_matches_fail(self):
        self._check_api_no_matches_fail(check_api3, 3)
