# Copyright 2013 Canonical Ltd.  This software is licensed under the GNU
# Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import copy
from datetime import (
    timedelta,
)
from charmworld.reports import (
    AccumulatorTimeSeries,
    AccumulatorCategorySeries,
    CountTimeSeries,
    DEFAULT_START,
    series_charm_growth,
    TimeSeries,
)
from charmworld.testing import (
    factory,
    MongoTestBase,
)


class TestReports(MongoTestBase):

    def test_series_charm_growth_reviewed(self):
        factory.makeCharm(self.db, promulgated=True)
        factory.makeCharm(self.db, promulgated=False)
        series = series_charm_growth(self.db)
        last_record = series.records()['series'][-1]
        self.assertEqual(1, last_record['ppa'])
        self.assertEqual(1, last_record['distro'])


def getValues(series):
    serialized = series.serialize()
    values = [v for (k, v) in serialized]
    return values


class TestTimeSeries(MongoTestBase):

    def test_serialized(self):
        expected_items = 10
        step = timedelta(days=1)
        start = DEFAULT_START
        end = start + (expected_items * step)
        ts = TimeSeries(start, step, end)
        self.assertEqual(expected_items, len(ts._series))
        self.assertEqual(expected_items, len(ts.serialize()))
        expected = [0] * expected_items
        self.assertEqual(expected, getValues(ts))

    def test_nearest(self):
        step = timedelta(hours=8)
        start = DEFAULT_START
        end = start + timedelta(days=10)
        ts = TimeSeries(start, step, end)
        nearest = ts._nearest(start)
        self.assertEqual(nearest, start)
        sorted_series = sorted(ts._series)
        index = 1
        nearest = ts._nearest(start + step * index)
        expected = sorted_series[index]
        self.assertEqual(expected, nearest)
        index = 8
        nearest = ts._nearest(start + step * index + timedelta(hours=1))
        expected = sorted_series[index]
        self.assertEqual(expected, nearest)


class TestCountTimeSeries(MongoTestBase):

    def test_increment(self):
        expected_items = 10
        start = DEFAULT_START
        step = timedelta(days=1)
        end = start + (expected_items * step)
        ts = CountTimeSeries(start, step, end)
        third = ts.serialize()[3][0]
        ts.increment(third)
        expected = [0] * expected_items
        expected[3] = 1
        self.assertEqual(expected, getValues(ts))
        expected[3] = 2
        ts.increment(third)
        self.assertEqual(expected, getValues(ts))

    def test_increment_with_invalid_time(self):
        n_steps = 10
        start = DEFAULT_START
        step = timedelta(days=1)
        end = start + (n_steps * step)
        too_early = start - timedelta(seconds=1)
        ts = CountTimeSeries(start, step, end)
        ts.increment(too_early)
        expected = [(start + step * i, 0) for i in range(n_steps)]
        self.assertEqual(expected, ts.serialize())
        ts = CountTimeSeries(start, step, end)
        ts.increment(end)
        self.assertEqual(expected, ts.serialize())


def _bump(original, index):
    # Make a copy of the incoming list.
    mod = original[:]
    for i in range(index, len(original)):
        mod[i] += 1
    return mod


class TestAccumulatorTimeSeries(MongoTestBase):

    def test_increment(self):
        expected_items = 10
        start = DEFAULT_START
        step = timedelta(days=1)
        end = start + (expected_items * step)
        ts = AccumulatorTimeSeries(start, step, end)
        index = 3
        target = ts.serialize()[index][0]
        # Increment the list.
        ts.increment(target, 1)
        expected = [0] * expected_items
        expected = _bump(expected, index)
        self.assertEqual(expected, getValues(ts))
        # Since the value passed is the same nothing will be incremented.
        ts.increment(target, 1)
        self.assertEqual(expected, getValues(ts))
        # And this call increments again, using a different point.
        index = 6
        target = ts.serialize()[index][0]
        ts.increment(target, 2)
        expected = _bump(expected, index)
        self.assertEqual(expected, getValues(ts))
        self.assertEqual([1, 2], ts.records()['values'])
        ts.increment(target, 0)
        self.assertEqual([0, 1, 2], ts.records()['values'])

    def test_increment_with_invalid_time(self):
        # Input data with timestamps outside of the interval
        # [_start_date, _end_date + _step] is ignored.
        n_steps = 10
        start = DEFAULT_START
        step = timedelta(days=1)
        end = start + (n_steps * step)
        too_early = start - timedelta(seconds=1)
        ts = AccumulatorTimeSeries(start, step, end)
        ts.increment(too_early, 1)
        expected = [(start + step * i, 0) for i in range(n_steps)]
        self.assertEqual(expected, ts.serialize())
        ts = AccumulatorTimeSeries(start, step, end)
        ts.increment(end, 2)
        self.assertEqual(expected, ts.serialize())


def _bumpCategory(original, index, category):
    # Make a copy of the incoming list.
    mod = copy.deepcopy(original)
    for i in range(index, len(original)):
        if category not in mod[i]:
            mod[i][category] = 0
        mod[i][category] += 1
    return mod


class TestAccumulatorCategorySeries(MongoTestBase):

    def test_increment(self):
        expected_items = 10
        start = DEFAULT_START
        step = timedelta(days=1)
        end = start + (expected_items * step)
        ts = AccumulatorCategorySeries(start, step, end)
        self.assertEqual([{}] * expected_items, getValues(ts))

        index = 3
        target = ts.serialize()[index][0]
        # Increment the list.
        category = 'a'
        ts.increment(target, 1, category)
        expected = [dict() for x in range(expected_items)]
        expected = _bumpCategory(expected, index, category)
        self.assertEqual(expected, getValues(ts))

        # Since the (value, category) passed is the same nothing will be
        # incremented.
        ts.increment(target, 1, category)
        self.assertEqual(expected, getValues(ts))

        # And this call increments again, using a different point.
        index = 6
        target = ts.serialize()[index][0]
        ts.increment(target, 2, category)
        expected = _bumpCategory(expected, index, category)
        self.assertEqual(expected, getValues(ts))

        # And this call increments because a different category is used.
        category2 = 'b'
        index = 8
        target = ts.serialize()[index][0]
        ts.increment(target, 2, category2)
        expected = _bumpCategory(expected, index, category2)
        self.assertEqual(expected, getValues(ts))

        records = ts.records()['series']
        for r in records[0:3]:
            self.assertIn('timespan', r)
            self.assertNotIn(category, r)
            self.assertNotIn(category2, r)
        for r in records[3:7]:
            self.assertIn('timespan', r)
            self.assertIn(category, r)
            self.assertNotIn(category2, r)
        for r in records[8:]:
            self.assertIn('timespan', r)
            self.assertIn(category, r)
            self.assertIn(category2, r)

    def test_increment_with_invalid_time(self):
        # Input data with timestamps outside of the interval
        # [_start_date, _end_date + _step) is ignored.
        n_steps = 10
        start = DEFAULT_START
        step = timedelta(days=1)
        end = start + (n_steps * step)
        too_early = start - timedelta(seconds=1)
        ts = AccumulatorCategorySeries(start, step, end)
        ts.increment(too_early, 1, 'foo')
        expected = [(start + step * i, {}) for i in range(n_steps)]
        self.assertEqual(expected, ts.serialize())

        ts = AccumulatorCategorySeries(start, step, end)
        ts.increment(end, 1, 'foo')
        self.assertEqual(expected, ts.serialize())
