# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld import cached_view_config
from charmworld.models import Bundle
from charmworld.views.api import json_response


class BundleDetail(Bundle):

    service_keys = (
        'branch',
        'charm',
        'constraints',
        'num_units',
        'options',
    )

    def __init__(self, bundle_data):
        super(BundleDetail, self).__init__(bundle_data)
        data = self._augmented = {}
        data['services_list'] = svclist = []

        # Flesh out the services and provide a default if there is none.
        services = self.data['services'] or {}
        for k, v in services.items():
            d = dict(name=k)
            for key in BundleDetail.service_keys:
                if key != 'options':
                    d[key] = v.get(key, u'')

            # Flatten options.
            options = v.get('options', {})
            opt_list = []
            for opt_key, opt_val in options.items():
                opt_list.append(
                    dict(name=opt_key,
                         value=opt_val))
            d['options'] = opt_list
            svclist.append(d)

        data['relations_list'] = rellist = []

        # Convert relations to a dict with 'left' and 'right' elements,
        # ensuring the right is a list.
        relations = self.data['relations'] or []
        for rel in relations:
            right = rel[1]
            if isinstance(right, basestring):
                right = [right]
            d = dict(left=rel[0],
                     right=', '.join(right))
            rellist.append(d)

    def __getattr__(self, name):
        try:
            return super(BundleDetail, self).__getattr__(name)
        except AttributeError:
            if name in self._augmented:
                return self._augmented[name]
            else:
                raise


def find_bundle(request, promulgated=False):
    spec = dict(
        name=request.matchdict['bundle'],
    )
    if promulgated:
        spec["promulgated"] = True
    else:
        spec["owner"] = request.matchdict['owner']
    return Bundle.from_query(spec, request.db)


def _bundle_view(request, bundle):
    bundle = BundleDetail(dict(bundle))
    is_owner = bool((request.user is not None) and
                    (request.user.nickname == bundle.owner))

    return dict(
        bundle=bundle,
        icon_path=None,
        is_owner=is_owner,
    )


@cached_view_config(
    route_name="personal-bundle",
    renderer="charmworld:templates/bundle.pt")
def personal_bundle(request):
    return _bundle_view(request, find_bundle(request))


@cached_view_config(
    route_name="personal-bundle-json")
def personal_bundle_json(request):
    bundle = find_bundle(request)
    if bundle is None:
        status = 404
        id_ = Bundle.construct_id(request.matchdict['owner'],
                                  request.matchdict['basket'],
                                  request.matchdict['bundle'])
        value = {'type': 'no_such_bundle', 'bundle_id': id_}
    else:
        status = 200
        value = {bundle.name: bundle.data}
    return json_response(status, value)
