# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.models import Bundle
from charmworld.testing import (
    factory,
    TestCase,
    ViewTestBase,
)
from charmworld.views.bundles import (
    BundleDetail,
    find_bundle,
    personal_bundle_json,
)


class TestBundleDetail(TestCase):

    @staticmethod
    def make_service_list(name):
        keys = sorted(BundleDetail.service_keys)
        svcs = {name: dict(zip(keys, range(len(keys))))}
        svcs[name]['options'] = {}
        return svcs

    def test_empty_lists(self):
        data = factory.get_bundle_data()
        bd = BundleDetail(data)
        self.assertEqual({}, bd.data['services'])
        self.assertEqual({}, bd.data['relations'])

    def test_flattened_services(self):
        services = self.make_service_list('svc1')
        data = factory.get_bundle_data(services=services)
        bd = BundleDetail(data)
        self.assertEqual(services, bd.data['services'])
        expected = [dict(name='svc1',
                         branch=0,
                         charm=1,
                         constraints=2,
                         num_units=3,
                         options=[])]
        self.assertEqual(expected, bd.services_list)

    def test_flattened_services_with_options(self):
        services = self.make_service_list('svc1')
        services['svc1']['options'] = dict(a=1, b=2)
        data = factory.get_bundle_data(services=services)

        bd = BundleDetail(data)
        self.assertEqual(services, bd.data['services'])
        expected = [dict(name='svc1',
                         branch=0,
                         charm=1,
                         constraints=2,
                         num_units=3,
                         options=[
                             dict(name='a', value=1),
                             dict(name='b', value=2),
                         ])
                    ]
        self.assertEqual(expected, bd.services_list)

    def test_flattened_relations(self):
        rels = [['a', 'b'],
                ['c', ['d', 'e']]]
        data = factory.get_bundle_data(relations=rels)

        bd = BundleDetail(data)
        self.assertEqual(rels, bd.data['relations'])

        expected = [dict(left='a',
                         right='b'),
                    dict(left='c',
                         right='d, e'),
                    ]
        self.assertEqual(expected, bd.relations_list)


class TestBundles(ViewTestBase):

    def test_found_bundle(self):
        bundle, bundle_data = factory.makeBundle(self.db)
        request = self.getRequest()
        request.matchdict = dict(bundle=bundle.name,
                                 owner=bundle.owner)
        found = find_bundle(request)
        self.assertEqual(bundle, found)

    def test_found_bundle_promulgated(self):
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
        request = self.getRequest()
        request.matchdict = dict(bundle=bundle.name)
        found = find_bundle(request, promulgated=True)
        self.assertEqual(bundle, found)

    def test_not_found_bundle(self):
        bundle, bundle_data = factory.makeBundle(self.db)
        request = self.getRequest()
        request.matchdict = dict(bundle='nobody-knows',
                                 owner=bundle.owner)
        found = find_bundle(request)
        self.assertEqual(None, found)

    def test_personal_charm_json(self):
        self.enable_routes()
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna', basket='ball/4',
            series='raring', promulgated=False,
        )
        request = self.getRequest()
        request.matchdict = dict(
            owner=bundle.owner,
            basket=bundle.basket,
            bundle=bundle.name,
        )
        expected = {'gigis': dict(
            series='raring',
            relations={},
            services={},
        )}
        response = personal_bundle_json(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(expected, response.json_body)

    def test_personal_charm_json_not_found(self):
        self.enable_routes()
        owner = 'no-one'
        basket = 'no-basket'
        bundle = 'no-name'
        request = self.getRequest()
        request.matchdict = dict(
            owner=owner,
            basket=basket,
            bundle=bundle,
        )
        response = personal_bundle_json(request)
        self.assertEqual(response.status_code, 404)
        id_ = Bundle.construct_id(owner, basket, bundle)
        expected = {'type': 'no_such_bundle', 'bundle_id': id_}
        self.assertEqual(expected, response.json_body)
