# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

from xml.etree.ElementTree import tostring

from charmworld.models import Charm
from charmworld.testing import factory
from charmworld.testing import ViewTestBase
from charmworld.testing import WebTestBase
from charmworld.testing.html import parse
from charmworld.views.feeds import (
    _charm_changes_feed,
    charm_changes,
    personal_charm_changes,
    recently_changed_feed,
)
from charmworld.views.helpers import format_change


class CharmChangesTestCase(ViewTestBase):

    def test_atom(self):
        # The view retrieves promulgated charms.
        ignore, charm_data = factory.makeCharm(
            self.db, owner='joe', name='ping', promulgated=True)
        # The view ignores unpromulgated charms,
        factory.makeCharm(self.db, name='ping', promulgated=False)
        request = self.getRequest()
        request.matchdict = {'charm': 'ping', 'series': 'precise'}
        charm = Charm(charm_data)
        changes = _charm_changes_feed(charm, request).writeString("utf8")
        response = charm_changes(request)
        self.assertEqual(changes, response.ubody)


class PersonalCharmChangesTestCase(ViewTestBase):

    def test_atom(self):
        # The view retrieves personal charms matching owner, series, and name.
        ignore, charm_data = factory.makeCharm(
            self.db, owner='joe', series='series', name='ping')
        # The view ignores charms from other users and series,
        factory.makeCharm(self.db, owner='other', series='series', name='ping')
        factory.makeCharm(self.db, owner='joe', series='other', name='ping')
        factory.makeCharm(self.db, owner='joe', series='series', name='other')
        request = self.getRequest()
        request.matchdict = {
            'charm': 'ping', 'series': 'series', 'owner': 'joe'}
        charm = Charm(charm_data)
        changes = _charm_changes_feed(charm, request).writeString("utf8")
        response = personal_charm_changes(request)
        self.assertEqual(changes, response.ubody)


class RecentlyChangedFeedTestCase(ViewTestBase):

    def test_atom(self):
        # The view retrieves charms and formats them as atom.
        ignore, charm_data_1 = factory.makeCharm(
            self.db, name='foo', commit_message='first\ncommit')
        ignore, charm_data_2 = factory.makeCharm(self.db, name='bar')
        request = self.getRequest()
        response = recently_changed_feed(request)
        # Remove the namespace so that the etree queries can use the tags.
        tree = parse(
            response.ubody.replace('xmlns="http://www.w3.org/2005/Atom"', ''))
        root = tree.getroot()
        self.assertEqual('feed', root.tag)
        entries = root.findall('entry')
        self.assertEqual(2, len(entries))
        charm = Charm(charm_data_1)
        entry = entries[0]
        self.assertEqual('foo', entry.find('title').text)
        self.assertEqual('John Doe', entry.find('author/name').text)
        self.assertEqual('first<br/>commit', entry.find('summary').text)
        link = u"https://bazaar.launchpad.net/%s/revision/%s" % (
            charm.branch_spec.replace('~', '%7E'), charm.last_change["revno"])
        self.assertEqual(link, entry.find('link').attrib['href'])
        charm = Charm(charm_data_2)
        entry = entries[1]
        self.assertEqual('bar', entry.find('title').text)


class RecentlyChangedTestCase(WebTestBase):

    def test_html(self):
        # The view retrieves charms, formats them as html, but includes
        # a an atom link. This test complements
        # RecentlyChangedFeedTestCase.test_atom.
        ignore, charm_data_1 = factory.makeCharm(
            self.db, name='foo', commit_message='first\ncommit')
        ignore, charm_data_2 = factory.makeCharm(self.db, name='bar')
        response = self.app.get('/recently-changed')
        tree = parse(
            response.ubody.replace('xmlns="http://www.w3.org/2005/Atom"', ''))
        root = tree.getroot()
        atom_link = root.find('head/link[@type="application/atom+xml"]')
        self.assertEqual('/feed/charm-changes', atom_link.attrib['href'])
        rows = root.findall('body/*/table/tr')
        self.assertEqual(2, len(rows))
        charm = Charm(charm_data_1)
        cells = rows[0].findall('td')
        anchor = cells[0].find('a')
        self.assertEqual(charm.store_url, anchor.text)
        self.assertEqual(charm.short_url, anchor.attrib['href'])
        self.assertEqual(charm.summary, cells[1].text)
        self.assertEqual(
            '<td>%s</td>' % format_change(charm.last_change, charm),
            tostring(cells[2]).strip())
