# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.models import Charm
from charmworld.views.helpers import (
    format_change,
    interfaces,
    result_sorter,
)
from charmworld.testing import factory
from charmworld.testing import MongoTestBase


class TestHelpers(MongoTestBase):

    def test_interfaces(self):
        # view.helpers.interfaces() shows which interface are
        # available and which charms require and provide these
        # interfaces.
        one_id, one = factory.makeCharm(
            self.db, name='some_webapp',
            provides={'foo': {'interface': 'http'}},
            requires={'database': {'interface': 'pgsql'}},
            promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, name='some_dbserver',
            provides={'database': {'interface': 'pgsql'}}, requires={},
            promulgated=True)
        result = interfaces(self.db)

        # interfaces() does not retrieve the charm's series from the
        # Mongo DB, but Charm.__eq__ compares name, owner, series, so
        # we must delete the series from the expected data.
        del one['series']
        del two['series']
        one = Charm(one)
        two = Charm(two)
        expected = [
            {
                'name': 'http',
                'provided_by': [one],
                'required_by': [],
            },
            {
                'name': 'pgsql',
                'provided_by': [two],
                'required_by': [one],
            }
        ]
        self.assertEqual(expected, result)

    def test_interfaces_does_not_list_invalid_charms(self):
        # Invalid charms are not returned by interfaces().
        one_id, one = factory.makeCharm(
            self.db, name='some_webapp',
            provides={'foo': {'interface': 'http'}},
            requires={'database': {'interface': 'pgsql'}},
            promulgated=True)
        two_id, two = factory.makeCharm(
            self.db, name='some_dbserver',
            provides={'database': {'interface': 'pgsql'}}, requires={},
            charm_error=True, promulgated=True)
        result = interfaces(self.db)

        del one['series']
        del two['series']
        one = Charm(one)
        two = Charm(two)

        expected = [
            {
                'name': 'http',
                'provided_by': [one],
                'required_by': [],
            },
            {
                'name': 'pgsql',
                'provided_by': [],
                'required_by': [one],
            }
        ]
        self.assertEqual(expected, result)

    def test_interfaces_unreviewed_charm(self):
        # Unreviewed charms are only returned if no reviewed charm
        # with the same name exists.
        one_id, one = factory.makeCharm(
            self.db, name='some_dbserver', owner='foo',
            provides={'database': {'interface': 'pgsql'}}, requires={})
        result = interfaces(self.db)

        del one['series']
        one = Charm(one)
        expected = [
            {
                'name': 'pgsql',
                'provided_by': [one],
                'required_by': [],
            }
        ]
        self.assertEqual(expected, result)

        # If a reviewed charm with the same name exists, this charm is
        # listed instead.
        two_id, two = factory.makeCharm(
            self.db, name='some_dbserver',
            provides={'database': {'interface': 'pgsql'}}, requires={},
            promulgated=True)
        result = interfaces(self.db)

        del two['series']
        two = Charm(two)
        expected = [
            {
                'name': 'pgsql',
                'provided_by': [two],
                'required_by': [],
            }
        ]
        self.assertEqual(expected, result)

    def test_result_sorter_raring(self):
        charm_a = factory.get_charm_json(series='precise')
        charm_b = factory.get_charm_json(series='raring')
        with self.assertNoException():
            result_sorter(charm_a, charm_b)

    def test_result_sorter_saucy(self):
        charm_a = factory.get_charm_json(series='saucy')
        charm_b = factory.get_charm_json(series='raring')
        with self.assertNoException():
            result_sorter(charm_a, charm_b)

    def test_format_change(self):
        charm = Charm(factory.get_charm_json())
        result = format_change(charm.last_change)
        self.assertEqual('2012/07/23 John Doe maintainer', result)

    def test_format_change_empty_branch(self):
        charm = Charm(factory.get_charm_json(empty_branch=True))
        result = format_change(charm.last_change)
        self.assertEqual(' ', result)
