# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime

from charmworld.testing import ViewTestBase
from charmworld.testing import factory
from charmworld.views.reports import json_charm_changes
from charmworld.views.reports import json_contributor_growth
from charmworld.views.reports import json_charm_growth


class TestReportViews(ViewTestBase):

    def setUp(self):
        super(TestReportViews, self).setUp()
        self.creation_date = factory.random_date()
        change = factory.makeChange(created=self.creation_date)
        factory.makeCharm(self.db, changes=[change], promulgated=True)

    def _generateReport(self, report, key, is_accumulator):
        req = self.getRequest()
        rep = report(req)
        series = rep['series']
        # Series dates are based on a "nearest" algorithm to the creation
        # date, so we have to find what date in the series is nearest to our
        # creation date.
        true_date = lambda x: datetime.strptime(x, '%m/%d/%y')
        series_dates = reversed(
            sorted([true_date(s['timespan']) for s in series]))
        for series_date in series_dates:
            if self.creation_date >= series_date.date():
                break

        # If the series powering the report is an accumulator, we need all
        # entries starting from the series date on. Otherwise we just need
        # entries on the series date.
        self.assertLessEqual(series_date.date(), self.creation_date)
        if is_accumulator:
            entries = [s for s in series
                       if true_date(s['timespan']) >= series_date]
        else:
            entries = [s for s in series
                       if true_date(s['timespan']) == series_date]
        empty = [s for s in series if s not in entries]

        return entries, empty

    def assertReport(self, report, key, is_accumulator=True):
        # Asserts that the key value is changed appropriately at the start
        # date of the charm change.
        entries, empty = self._generateReport(report, key, is_accumulator)
        # All entries corresponding to the change are updated.
        expected = 1
        for entry in entries:
            self.assertEqual(entry[key], expected, "for report {}".format(key))
        # All entries not corresponding to the change (i.e. with no data) are
        # 0.
        expected = 0
        for entry in empty:
            self.assertEqual(entry[key], expected, "for report {}".format(key))

    def test_json_charm_changes(self):
        # The charm changes report correctly captures that a charm has
        # changed.
        self.assertReport(json_charm_changes, 'value', is_accumulator=False)

    def test_json_contributor_growth(self):
        # Test that the contributor report captures the increase in community
        # from a new charm.
        self.assertReport(json_contributor_growth, 'community')

    def test_json_charm_growth(self):
        # Test that the growth report captures the increase of a new charm.
        self.assertReport(json_charm_growth, 'distro')
