# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mock import patch
from pyelasticsearch import ElasticSearch

from charmworld.search import (
    ElasticSearchClient,
    IndexNotReady,
)
from charmworld.testing import (
    factory,
    ViewTestBase,
)
from charmworld.views import search


RETURN_VAL = {
    'results': [],
    'search_time': 0.0,
    'result_total': 0,
    'matches': 0,
}


class SearchTests(ViewTestBase):

    @patch('charmworld.views.search.do_search',
           autospec=True, return_value=RETURN_VAL)
    def test_search(self, mock):
        terms = {"search_text": 'foo'}
        keys = ['result_total', 'results', 'search_time']
        request = self.getRequest(params=terms)
        response = search.search(request)

        mock.assert_called_with(request)
        self.assertEqual(keys, sorted(response.keys()))
        self.assertEqual(0, response['result_total'])
        self.assertEqual([], response['results'])
        self.assertEqual(0.0, response['search_time'])

    @patch('charmworld.views.search.do_search',
           autospec=True, side_effect=IndexNotReady)
    def test_search_503(self, mock):
        terms = {"search_text": 'foo'}
        request = self.getRequest(params=terms)
        # render_to_response requires the request to be much more like a real
        # request, with scheme and static stuff handling, so mock it out.
        with patch('charmworld.views.search.render_to_response') as render:
            response = search.search(request)
        render.assert_called_with('../templates/search-not-ready.pt', {},
                                  request=request)
        self.assertEqual(503, response.status_code)

    def test_search_503_on_connection_error(self):
        terms = {"search_text": 'foo'}
        request = self.getRequest(params=terms)
        request.index_client = ElasticSearchClient(
            ElasticSearch(['http://localhost:70']), 'foo')
        # render_to_response requires the request to be much more like a real
        # request, with scheme and static stuff handling, so mock it out.
        with patch('charmworld.views.search.render_to_response') as render:
            response = search.search(request)
        render.assert_called_with('../templates/search-not-ready.pt', {},
                                  request=request)
        self.assertEqual(503, response.status_code)

    def test_search_json(self):
        # The search json view substitutes the search result object with the
        # charms and adds some metadata.
        charm = factory.makeCharm(self.db, name='foo')[1]
        factory.makeCharm(self.db)
        self.use_index_client()
        self.index_client.index_charm(charm)
        terms = {"search_text": 'foo'}
        request = self.getRequest(params=terms)
        keys = sorted([
            'charm_total',
            'matches',
            'matches_human_readable_estimate',
            'results',
            'results_size',
            'search_time'])
        json = search.search_json(request).json
        self.assertEqual(keys, sorted(json.keys()))
        self.assertEqual(1, json['charm_total'])
        self.assertEqual(1, json['matches'])
        self.assertEqual('1', json['matches_human_readable_estimate'])
        self.assertEqual(1, json['results_size'])
        self.assertEqual(1, len(json['results']))
        found = json['results'][0]
        self.assertEqual('foo', found['name'])
        self.assertEqual(charm['series'], found['series'])
        self.assertEqual(charm['owner'], found['owner'])

    @patch('charmworld.views.search.do_search',
           autospec=True, side_effect=IndexNotReady)
    def test_search_json_503(self, mock):
        terms = {"search_text": 'foo'}
        request = self.getRequest(params=terms)
        response = search.search_json(request)
        self.assertEqual(503, response.status_code)
        self.assertEqual({'type': 'search_not_ready'}, response.json)

    def test_search_json_503_on_connection_error(self):
        terms = {"search_text": 'foo'}
        request = self.getRequest(params=terms)
        request.index_client = ElasticSearchClient(
            ElasticSearch(['http://localhost:70']), 'foo')
        response = search.search_json(request)
        self.assertEqual(503, response.status_code)
        self.assertEqual({'type': 'search_not_ready'}, response.json)
