# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import datetime

from charmworld.models import Charm
from charmworld.views.tools import (
    charm_store_missing,
    proof_errors,
)
from charmworld.testing import factory
from charmworld.testing import ViewTestBase
from charmworld.testing import WebTestBase
from charmworld.testing.html import (
    find_by_id,
    parse,
)


class TestReviews(WebTestBase):

    def test_last_modified_formatting(self):
        self.db.review_queue.save(
            factory.make_review_entry(date_modified=datetime(2012, 1, 2)))
        self.db.review_queue.save(
            factory.make_review_entry(date_modified=datetime(2012, 1, 3)))
        resp = self.app.get('/review-queue')
        tree = parse(resp.body)
        table = find_by_id(tree, 'review-table')
        headings = [th.text for th in table.findall('thead/tr/th')]
        self.assertEqual(
            ['Age', 'Date Created', 'Last Modified', 'Summary', 'Status'],
            headings)
        rows = table.findall('tr')
        self.assertEqual(2, len(rows))
        colmap = {}
        for row in rows:
            columns = row.findall('td')
            self.assertEqual(len(headings), len(columns))
            [colmap.setdefault(heading, []).append(col.text)
             for heading, col in zip(headings, columns)]
        self.assertEqual(['2012/01/02', '2012/01/03'], colmap['Last Modified'])


class CharmStoreMissingTestCase(ViewTestBase):

    def test_charms_with_store_errors(self):
        ignore, charm_1 = factory.makeCharm(self.db, charm_error=True)
        factory.makeCharm(self.db, charm_error=False)
        response = charm_store_missing(self.getRequest())
        charm = Charm(charm_1)
        self.assertEqual([charm], response['queue'])

    def test_obsolete_charms_ignored(self):
        ignore, charm_1 = factory.makeCharm(self.db, charm_error=True,
                                            series='oneiric')
        response = charm_store_missing(self.getRequest())
        self.assertEqual([], response['queue'])


class ProofErrorsTestCase(ViewTestBase):

    def test_promulgated_charms_with_errors(self):
        proof_with_errors = {'e': ['error 1']}
        proof_with_warnings = {'w': ['warning 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors)
        # The view ignores promulgated charms with proof warning.
        factory.makeCharm(self.db, promulgated=True, proof=proof_with_warnings)
        # The view ignored unpromulgated charms with proof errors.
        factory.makeCharm(self.db, promulgated=False, proof=proof_with_errors)
        response = proof_errors(self.getRequest())
        charm = Charm(charm_1)
        self.assertEqual([charm], response['charms'])

    def test_obsolete_charms_ignored(self):
        proof_with_errors = {'e': ['error 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors,
            series='oneiric')
        response = proof_errors(self.getRequest())
        self.assertEqual([], response['charms'])
